# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property
from typing import Any
from typing import List

from earthkit.data import FieldList

from ...dates.groups import GroupOfDates
from .action import Action
from .action import action_factory
from .empty import EmptyResult
from .misc import _tidy
from .misc import assert_fieldlist
from .result import Result
from .template import notify_result
from .trace import trace_datasource
from .trace import trace_select

LOG = logging.getLogger(__name__)


class JoinResult(Result):
    """Represents a result that combines multiple results.

    Attributes
    ----------
    context : object
        The context object.
    action_path : list
        The action path.
    group_of_dates : GroupOfDates
        The group of dates.
    results : List[Result]
        The list of results.
    """

    def __init__(
        self, context: object, action_path: list, group_of_dates: GroupOfDates, results: List[Result], **kwargs: Any
    ) -> None:
        """Initializes a JoinResult instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : list
            The action path.
        group_of_dates : GroupOfDates
            The group of dates.
        results : List[Result]
            The list of results.
        """
        super().__init__(context, action_path, group_of_dates)
        self.results: List[Result] = [r for r in results if not r.empty]

    @cached_property
    @assert_fieldlist
    @notify_result
    @trace_datasource
    def datasource(self) -> FieldList:
        """Returns the combined datasource from all results."""
        ds: FieldList = EmptyResult(self.context, self.action_path, self.group_of_dates).datasource
        for i in self.results:
            ds += i.datasource
        return _tidy(ds)

    def __repr__(self) -> str:
        """Returns a string representation of the JoinResult instance."""
        content: str = "\n".join([str(i) for i in self.results])
        return self._repr(content)


class JoinAction(Action):
    """Represents an action that combines multiple actions.

    Attributes
    ----------
    context : object
        The context object.
    action_path : list
        The action path.
    actions : List[Action]
        The list of actions.
    """

    def __init__(self, context: object, action_path: list, *configs: dict) -> None:
        """Initializes a JoinAction instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : list
            The action path.
        *configs : dict
            The configuration dictionaries.
        """
        super().__init__(context, action_path, *configs)
        self.actions: List[Action] = [action_factory(c, context, action_path + [str(i)]) for i, c in enumerate(configs)]

    def __repr__(self) -> str:
        """Returns a string representation of the JoinAction instance."""
        content: str = "\n".join([str(i) for i in self.actions])
        return self._repr(content)

    @trace_select
    def select(self, group_of_dates: GroupOfDates) -> JoinResult:
        """Selects the results for the given group of dates.

        Parameters
        ----------
        group_of_dates : GroupOfDates
            The group of dates.

        Returns
        -------
        JoinResult
            The combined result for the given group of dates.
        """
        results: List[Result] = [a.select(group_of_dates) for a in self.actions]
        return JoinResult(self.context, self.action_path, group_of_dates, results)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from collections import defaultdict
from typing import Any
from typing import Dict

import earthkit.data as ekd
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(context: Any, input: ekd.FieldList, orog: str, z: str = "z") -> ekd.FieldList:
    """Convert orography [m] to z (geopotential height).

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    input : FieldList
        List of input fields.
    orog : str
        Orography parameter.
    z : str, optional
        Geopotential height parameter. Defaults to "z".

    Returns
    -------
    FieldList
        List of fields with geopotential height.
    """
    result = []
    processed_fields: Dict[tuple, Dict[str, Any]] = defaultdict(dict)

    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")
        if param == orog:
            key = tuple(key.items())

            if param in processed_fields[key]:
                raise ValueError(f"Duplicate field {param} for {key}")

            output = f.to_numpy(flatten=True) * 9.80665
            result.append(new_field_from_numpy(f, output, param=z))
        else:
            result.append(f)

    return new_fieldlist_from_list(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import re
from typing import Any
from typing import Dict
from typing import Optional

import earthkit.data as ekd
from earthkit.data.indexing.fieldlist import FieldArray

from .legacy import legacy_filter


class RenamedFieldMapping:
    """Rename a field based on the value of another field.

    Parameters
    ----------
    field : Any
        The field to be renamed.
    what : str
        The name of the field that will be used to rename the field.
    renaming : dict of dict of str
        A dictionary mapping the values of 'what' to the new names.
    """

    def __init__(self, field: Any, what: str, renaming: Dict[str, Dict[str, str]]) -> None:
        """Initialize a RenamedFieldMapping instance.

        Parameters
        ----------
        field : Any
            The field to be renamed.
        what : str
            The name of the field that will be used to rename the field.
        renaming : dict of dict of str
            A dictionary mapping the values of 'what' to the new names.
        """
        self.field = field
        self.what = what
        self.renaming = {}
        for k, v in renaming.items():
            self.renaming[k] = {str(a): str(b) for a, b in v.items()}

    def metadata(self, key: Optional[str] = None, **kwargs: Any) -> Any:
        """Get metadata from the original field, with the option to rename the parameter.

        Parameters
        ----------
        key : str, optional
            The metadata key.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Any
            The metadata value.
        """
        if key is None:
            return self.field.metadata(**kwargs)

        value = self.field.metadata(key, **kwargs)
        if key == self.what:
            return self.renaming.get(self.what, {}).get(value, value)

        return value

    def __getattr__(self, name: str) -> Any:
        """Get an attribute from the original field.

        Parameters
        ----------
        name : str
            The name of the attribute.

        Returns
        -------
        Any
            The attribute value.
        """
        return getattr(self.field, name)

    def __repr__(self) -> str:
        """Get the string representation of the original field.

        Returns
        -------
        str
            The string representation of the original field.
        """
        return repr(self.field)


class RenamedFieldFormat:
    """Rename a field based on a format string.

    Parameters
    ----------
    field : Any
        The field to be renamed.
    what : str
        The name of the field that will be used to rename the field.
    format : str
        The format string for renaming.
    """

    def __init__(self, field: Any, what: str, format: str) -> None:
        """Initialize a RenamedFieldFormat instance.

        Parameters
        ----------
        field : Any
            The field to be renamed.
        what : str
            The name of the field that will be used to rename the field.
        format : str
            The format string for renaming.
        """
        self.field = field
        self.what = what
        self.format = format
        self.bits = re.findall(r"{(\w+)}", format)

    def metadata(self, *args: Any, **kwargs: Any) -> Any:
        """Get metadata from the original field, with the option to rename the parameter using a format string.

        Parameters
        ----------
        *args : Any
            Positional arguments.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Any
            The metadata value.
        """
        value = self.field.metadata(*args, **kwargs)
        if args:
            assert len(args) == 1
            if args[0] == self.what:
                bits = {b: self.field.metadata(b, **kwargs) for b in self.bits}
                return self.format.format(**bits)
        return value

    def __getattr__(self, name: str) -> Any:
        """Get an attribute from the original field.

        Parameters
        ----------
        name : str
            The name of the attribute.

        Returns
        -------
        Any
            The attribute value.
        """
        return getattr(self.field, name)

    def __repr__(self) -> str:
        """Get the string representation of the original field.

        Returns
        -------
        str
            The string representation of the original field.
        """
        return repr(self.field)


@legacy_filter(__file__)
def execute(context: Any, input: ekd.FieldList, what: str = "param", **kwargs: Any) -> ekd.FieldList:
    """Rename fields based on the value of another field or a format string.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    input : List[Any]
        List of input fields.
    what : str, optional
        The field to be used for renaming. Defaults to "param".
    **kwargs : Any
        Additional keyword arguments for renaming.

    Returns
    -------
    ekd.FieldList
        Array of renamed fields.
    """

    if what in kwargs and isinstance(kwargs[what], str):
        return FieldArray([RenamedFieldFormat(fs, what, kwargs[what]) for fs in input])

    return FieldArray([RenamedFieldMapping(fs, what, kwargs) for fs in input])



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import inspect
import logging
import os
from typing import Any
from typing import Callable

from anemoi.datasets.create.input.template import resolve

from ..source import Source
from . import source_registry

LOG = logging.getLogger(__name__)


class LegacySource(Source):
    """A legacy source class.

    Parameters
    ----------
    context : Any
        The context in which the source is created.
    *args : tuple
        Positional arguments.
    **kwargs : dict
        Keyword arguments.
    """

    def __init__(self, context: Any, *args: Any, **kwargs: Any) -> None:
        super().__init__(context, *args, **kwargs)
        self.args = args
        self.kwargs = kwargs


class legacy_source:
    """A decorator class for legacy sources.

    Parameters
    ----------
    name : str
        The name of the legacy source.
    """

    def __init__(self, name: str) -> None:
        name, _ = os.path.splitext(os.path.basename(name))
        self.name = name

    def __call__(self, execute: Callable) -> Callable:
        """Call method to wrap the execute function.

        Parameters
        ----------
        execute : function
            The execute function to be wrapped.

        Returns
        -------
        function
            The wrapped execute function.
        """
        this = self
        name = f"Legacy{self.name.title()}Source"
        source = ".".join([execute.__module__, execute.__name__])

        def execute_wrapper(self, dates) -> Any:
            """Wrapper method to call the execute function."""

            args, kwargs = resolve(self.context, (self.args, self.kwargs))

            try:
                return execute(self.context, dates, *args, **kwargs)
            except TypeError:
                LOG.error(f"Error executing source {this.name} from {source}")
                LOG.error(f"Function signature is: {inspect.signature(execute)}")
                LOG.error(f"Arguments are: {args=}, {kwargs=}")
                raise

        klass = type(
            name,
            (LegacySource,),
            {
                "execute": execute_wrapper,
                "_source": source,
            },
        )

        source_registry.register(self.name)(klass)

        return execute



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.output import new_grib_output

from anemoi.datasets.create.utils import to_datetime_list

from .legacy import legacy_source


def _date_to_datetime(d: Any) -> Any:
    """Converts a date string or a list/tuple of date strings to datetime objects.

    Parameters
    ----------
    d : Any
        A date string or a list/tuple of date strings.

    Returns
    -------
    Any
        A datetime object or a list/tuple of datetime objects.
    """
    if isinstance(d, (list, tuple)):
        return [_date_to_datetime(x) for x in d]
    return datetime.datetime.fromisoformat(d)


def normalise_time_delta(t: Any) -> datetime.timedelta:
    """Normalizes a time delta string to a datetime.timedelta object.

    Parameters
    ----------
    t : Any
        A time delta string ending with 'h' or a datetime.timedelta object.

    Returns
    -------
    datetime.timedelta
        A normalized datetime.timedelta object.
    """
    if isinstance(t, datetime.timedelta):
        assert t == datetime.timedelta(hours=t.hours), t

    assert t.endswith("h"), t

    t = int(t[:-1])
    t = datetime.timedelta(hours=t)
    return t


def group_by_field(ds: Any) -> Dict[Tuple, List[Any]]:
    """Groups fields by their metadata excluding 'date', 'time', and 'step'.

    Parameters
    ----------
    ds : Any
        A dataset object.

    Returns
    -------
    Dict[Tuple, List[Any]]
        A dictionary where keys are tuples of metadata items and values are lists of fields.
    """
    d = defaultdict(list)
    for field in ds.order_by("valid_datetime"):
        m = field.metadata(namespace="mars")
        for k in ("date", "time", "step"):
            m.pop(k, None)
        keys = tuple(m.items())
        d[keys].append(field)
    return d


@legacy_source(__file__)
def tendencies(dates: List[datetime.datetime], time_increment: Any, **kwargs: Any) -> Any:
    """Computes tendencies for the given dates and time increment.

    Parameters
    ----------
    dates : List[datetime.datetime]
        A list of datetime objects.
    time_increment : Any
        A time increment string ending with 'h' or a datetime.timedelta object.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    Any
        A dataset object with computed tendencies.
    """
    print("✅", kwargs)
    time_increment = normalise_time_delta(time_increment)

    shifted_dates = [d - time_increment for d in dates]
    all_dates = sorted(list(set(dates + shifted_dates)))

    # from .mars import execute as mars
    from anemoi.datasets.create.mars import execute as mars

    ds = mars(dates=all_dates, **kwargs)

    dates_in_data = ds.unique_values("valid_datetime", progress_bar=False)["valid_datetime"]
    for d in all_dates:
        assert d.isoformat() in dates_in_data, d

    ds1 = ds.sel(valid_datetime=[d.isoformat() for d in dates])
    ds2 = ds.sel(valid_datetime=[d.isoformat() for d in shifted_dates])

    assert len(ds1) == len(ds2), (len(ds1), len(ds2))

    group1 = group_by_field(ds1)
    group2 = group_by_field(ds2)

    assert group1.keys() == group2.keys(), (group1.keys(), group2.keys())

    # prepare output tmp file so we can read it back
    tmp = temp_file()
    path = tmp.path
    out = new_grib_output(path)

    for k in group1:
        assert len(group1[k]) == len(group2[k]), k
        print()
        print("❌", k)

        for field, b_field in zip(group1[k], group2[k]):
            for k in ["param", "level", "number", "grid", "shape"]:
                assert field.metadata(k) == b_field.metadata(k), (
                    k,
                    field.metadata(k),
                    b_field.metadata(k),
                )

            c = field.to_numpy()
            b = b_field.to_numpy()
            assert c.shape == b.shape, (c.shape, b.shape)

            ################
            # Actual computation happens here
            x = c - b
            ################

            assert x.shape == c.shape, c.shape
            print(f"Computing data for {field.metadata('valid_datetime')}={field}-{b_field}")
            out.write(x, template=field)

    out.close()

    from earthkit.data import from_source

    ds = from_source("file", path)
    # save a reference to the tmp file so it is deleted
    # only when the dataset is not used anymore
    ds._tmp = tmp

    return ds


execute = tendencies

if __name__ == "__main__":
    import yaml

    config = yaml.safe_load(
        """

    config:
      time_increment: 12h
      database: marser
      class: ea
      # date: computed automatically
      # time: computed automatically
      expver: "0001"
      grid: 20.0/20.0
      levtype: sfc
      param: [2t]
    """
    )["config"]

    dates = yaml.safe_load("[2022-12-30 18:00, 2022-12-31 00:00, 2022-12-31 06:00, 2022-12-31 12:00]")
    dates = to_datetime_list(dates)

    DEBUG = True
    for f in tendencies(dates, **config):
        print(f, f.to_numpy().mean())



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""This module provides a function to convert u and v wind components to wind speed and direction."""

from collections import defaultdict
from typing import Any
from typing import List

import earthkit.data as ekd
import numpy as np
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list
from earthkit.meteo.wind.array import xy_to_polar

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(
    context: Any,
    input: List[Any],
    u_component: str,
    v_component: str,
    wind_speed: str,
    wind_dir: str,
    in_radians: bool = False,
) -> ekd.FieldList:
    """Converts u and v wind components to wind speed and direction.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    input : List[Any]
        List of input fields containing wind components.
    u_component : str
        The name of the u component field.
    v_component : str
        The name of the v component field.
    wind_speed : str
        The name of the wind speed field to be created.
    wind_dir : str
        The name of the wind direction field to be created.
    in_radians : bool, optional
        If True, the wind direction is returned in radians. Default is False.

    Returns
    -------
    ekd.FieldList
        A FieldArray containing the wind speed and direction fields.
    """
    result = []

    wind_params = (u_component, v_component)
    wind_pairs = defaultdict(dict)

    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")

        if param not in wind_params:
            result.append(f)
            continue

        key = tuple(key.items())

        if param in wind_pairs[key]:
            raise ValueError(f"Duplicate wind component {param} for {key}")

        wind_pairs[key][param] = f

    for _, pairs in wind_pairs.items():
        if len(pairs) != 2:
            raise ValueError("Missing wind component")

        u = pairs[u_component]
        v = pairs[v_component]

        # assert speed.grid_mapping == dir.grid_mapping
        magnitude, direction = xy_to_polar(u.to_numpy(flatten=True), v.to_numpy(flatten=True))
        if in_radians:
            direction = np.deg2rad(direction)

        result.append(new_field_from_numpy(u, magnitude, params=wind_speed))
        result.append(new_field_from_numpy(v, direction, params=wind_dir))

    return new_fieldlist_from_list(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np
from anemoi.transform.fields import new_field_with_valid_datetime
from anemoi.transform.fields import new_fieldlist_from_list
from anemoi.utils.dates import as_datetime
from anemoi.utils.dates import frequency_to_timedelta

from .action import Action
from .action import action_factory
from .join import JoinResult
from .result import Result
from .trace import trace_select

LOG = logging.getLogger(__name__)


class DateMapper:
    """A factory class to create DateMapper instances based on the given mode."""

    @staticmethod
    def from_mode(mode: str, source: Any, config: Dict[str, Any]) -> "DateMapper":
        """Create a DateMapper instance based on the given mode.

        Parameters
        ----------
        mode : str
            The mode to use for the DateMapper.
        source : Any
            The data source.
        config : dict
            Configuration parameters.

        Returns
        -------
        DateMapper
            An instance of DateMapper.
        """
        MODES: dict = dict(
            closest=DateMapperClosest,
            climatology=DateMapperClimatology,
            constant=DateMapperConstant,
        )

        if mode not in MODES:
            raise ValueError(f"Invalid mode for DateMapper: {mode}")

        return MODES[mode](source, **config)


class DateMapperClosest(DateMapper):
    """A DateMapper implementation that maps dates to the closest available dates."""

    def __init__(self, source: Any, frequency: str = "1h", maximum: str = "30d", skip_all_nans: bool = False) -> None:
        """Initialize DateMapperClosest.

        Parameters
        ----------
        source : Any
            The data source.
        frequency : str
            Frequency of the dates.
        maximum : str
            Maximum time delta.
        skip_all_nans : bool
            Whether to skip all NaN values.
        """
        self.source: Any = source
        self.maximum: Any = frequency_to_timedelta(maximum)
        self.frequency: Any = frequency_to_timedelta(frequency)
        self.skip_all_nans: bool = skip_all_nans
        self.tried: Set[Any] = set()
        self.found: Set[Any] = set()

    def transform(self, group_of_dates: Any) -> Generator[Tuple[Any, Any], None, None]:
        """Transform the group of dates to the closest available dates.

        Parameters
        ----------
        group_of_dates : Any
            The group of dates to transform.

        Returns
        -------
        Generator[Tuple[Any, Any], None, None]
            Transformed dates.
        """
        from anemoi.datasets.dates.groups import GroupOfDates

        asked_dates = list(group_of_dates)
        if not asked_dates:
            return []

        to_try = set()
        for date in asked_dates:
            start = date
            while start >= date - self.maximum:
                to_try.add(start)
                start -= self.frequency

            end = date
            while end <= date + self.maximum:
                to_try.add(end)
                end += self.frequency

        to_try = sorted(to_try - self.tried)
        info = {k: "no-data" for k in to_try}

        if not to_try:
            LOG.warning(f"No new dates to try for {group_of_dates} in {self.source}")
            # return []

        if to_try:
            result = self.source.select(
                GroupOfDates(
                    sorted(to_try),
                    group_of_dates.provider,
                    partial_ok=True,
                )
            )

            cnt = 0
            for f in result.datasource:
                cnt += 1
                # We could keep the fields in a dictionary, but we don't want to keep the fields in memory
                date = as_datetime(f.metadata("valid_datetime"))

                if self.skip_all_nans:
                    if np.isnan(f.to_numpy()).all():
                        LOG.warning(f"Skipping {date} because all values are NaN")
                        info[date] = "all-nans"
                        continue

                info[date] = "ok"
                self.found.add(date)

            if cnt == 0:
                raise ValueError(f"No data found for {group_of_dates} in {self.source}")

            self.tried.update(to_try)

        if not self.found:
            for k, v in info.items():
                LOG.warning(f"{k}: {v}")

            raise ValueError(f"No matching data found for {asked_dates} in {self.source}")

        new_dates = defaultdict(list)

        for date in asked_dates:
            best = None
            for found_date in sorted(self.found):
                delta = abs(date - found_date)
                # With < we prefer the first date
                # With <= we prefer the last date
                if best is None or delta <= best[0]:
                    best = delta, found_date
            new_dates[best[1]].append(date)

        for date, dates in new_dates.items():
            yield (
                GroupOfDates([date], group_of_dates.provider),
                GroupOfDates(dates, group_of_dates.provider),
            )


class DateMapperClimatology(DateMapper):
    """A DateMapper implementation that maps dates to specified climatology dates."""

    def __init__(self, source: Any, year: int, day: int, hour: Optional[int] = None) -> None:
        """Initialize DateMapperClimatology.

        Parameters
        ----------
        source : Any
            The data source.
        year : int
            The year to map to.
        day : int
            The day to map to.
        hour : Optional[int]
            The hour to map to.
        """
        self.year: int = year
        self.day: int = day
        self.hour: Optional[int] = hour

    def transform(self, group_of_dates: Any) -> Generator[Tuple[Any, Any], None, None]:
        """Transform the group of dates to the specified climatology dates.

        Parameters
        ----------
        group_of_dates : Any
            The group of dates to transform.

        Returns
        -------
        Generator[Tuple[Any, Any], None, None]
            Transformed dates.
        """
        from anemoi.datasets.dates.groups import GroupOfDates

        dates = list(group_of_dates)
        if not dates:
            return []

        new_dates = defaultdict(list)
        for date in dates:
            new_date = date.replace(year=self.year, day=self.day)
            if self.hour is not None:
                new_date = new_date.replace(hour=self.hour, minute=0, second=0)
            new_dates[new_date].append(date)

        for date, dates in new_dates.items():
            yield (
                GroupOfDates([date], group_of_dates.provider),
                GroupOfDates(dates, group_of_dates.provider),
            )


class DateMapperConstant(DateMapper):
    """A DateMapper implementation that maps dates to a constant date."""

    def __init__(self, source: Any, date: Optional[Any] = None) -> None:
        """Initialize DateMapperConstant.

        Parameters
        ----------
        source : Any
            The data source.
        date : Optional[Any]
            The constant date to map to.
        """
        self.source: Any = source
        self.date: Optional[Any] = date

    def transform(self, group_of_dates: Any) -> Tuple[Any, Any]:
        """Transform the group of dates to a constant date.

        Parameters
        ----------
        group_of_dates : Any
            The group of dates to transform.

        Returns
        -------
        Tuple[Any, Any]
            Transformed dates.
        """
        from anemoi.datasets.dates.groups import GroupOfDates

        if self.date is None:
            return [
                (
                    GroupOfDates([], group_of_dates.provider),
                    group_of_dates,
                )
            ]

        return [
            (
                GroupOfDates([self.date], group_of_dates.provider),
                group_of_dates,
            )
        ]


class DateMapperResult(Result):
    """A Result implementation that updates the valid datetime of the datasource."""

    def __init__(
        self,
        context: Any,
        action_path: List[str],
        group_of_dates: Any,
        source_result: Any,
        mapper: DateMapper,
        original_group_of_dates: Any,
    ) -> None:
        """Initialize DateMapperResult.

        Parameters
        ----------
        context : Any
            The context.
        action_path : list of str
            The action path.
        group_of_dates : Any
            The group of dates.
        source_result : Any
            The source result.
        mapper : DateMapper
            The date mapper.
        original_group_of_dates : Any
            The original group of dates.
        """
        super().__init__(context, action_path, group_of_dates)

        self.source_results: Any = source_result
        self.mapper: DateMapper = mapper
        self.original_group_of_dates: Any = original_group_of_dates

    @property
    def datasource(self) -> Any:
        """Get the datasource with updated valid datetime."""
        result: list = []

        for field in self.source_results.datasource:
            for date in self.original_group_of_dates:
                result.append(new_field_with_valid_datetime(field, date))

        if not result:
            raise ValueError("repeated_dates: no input data found")

        return new_fieldlist_from_list(result)


class RepeatedDatesAction(Action):
    """An Action implementation that selects and transforms a group of dates."""

    def __init__(self, context: Any, action_path: List[str], source: Any, mode: str, **kwargs: Any) -> None:
        """Initialize RepeatedDatesAction.

        Args:
            context (Any): The context.
            action_path (List[str]): The action path.
            source (Any): The data source.
            mode (str): The mode for date mapping.
            **kwargs (Any): Additional arguments.
        """
        super().__init__(context, action_path, source, mode, **kwargs)

        self.source: Any = action_factory(source, context, action_path + ["source"])
        self.mapper: DateMapper = DateMapper.from_mode(mode, self.source, kwargs)

    @trace_select
    def select(self, group_of_dates: Any) -> JoinResult:
        """Select and transform the group of dates.

        Args:
            group_of_dates (Any): The group of dates to select.

        Returns
        -------
        JoinResult
            The result of the join operation.
        """
        results: list = []
        for one_date_group, many_dates_group in self.mapper.transform(group_of_dates):
            results.append(
                DateMapperResult(
                    self.context,
                    self.action_path,
                    one_date_group,
                    self.source.select(one_date_group),
                    self.mapper,
                    many_dates_group,
                )
            )

        return JoinResult(self.context, self.action_path, group_of_dates, results)

    def __repr__(self) -> str:
        """Get the string representation of the action.

        Returns
        -------
        str
            The string representation.
        """
        return f"MultiDateMatchAction({self.source}, {self.mapper})"



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import Optional
from typing import Set

from numpy.typing import NDArray

from . import open_dataset
from .dataset import Dataset
from .debug import Node
from .forwards import Forwards

LOG = logging.getLogger(__name__)


class Statistics(Forwards):
    """A class to represent statistics for a dataset.

    Attributes
    ----------
    dataset : Dataset
        The dataset object.
    statistic : Any
        The statistic data.
    """

    def __init__(self, dataset: Dataset, statistic: Any) -> None:
        """Initialize the Statistics object.

        Parameters
        ----------
        dataset : Dataset
            The dataset object.
        statistic : Any
            The statistic data.
        """
        super().__init__(dataset)
        self._statistic = open_dataset(statistic, select=dataset.variables)
        # TODO: relax that check to allow for a subset of variables
        if dataset.variables != self._statistic.variables:
            raise ValueError(
                f"Incompatible variables: {dataset.variables} and {self._statistic.variables} ({dataset} {self._statistic})"
            )

    @cached_property
    def statistics(self) -> Dict[str, NDArray[Any]]:
        """Get the statistics."""
        return self._statistic.statistics

    def statistics_tendencies(self, delta: Optional[datetime.timedelta] = None) -> Dict[str, NDArray[Any]]:
        """Get the statistics tendencies.

        Parameters
        ----------
        delta : Optional[datetime.timedelta]
            The time delta.

        Returns
        -------
        Dict[str, NDArray[Any]]
            The statistics tendencies.
        """
        if delta is None:
            delta = self.frequency
        return self._statistic.statistics_tendencies(delta)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the forwards subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata specific to the forwards subclass.
        """
        return dict(statistics=self._statistic.metadata_specific())

    def tree(self) -> Node:
        """Get the tree representation of the statistics.

        Returns
        -------
        Node
            The tree representation of the statistics.
        """
        return Node(self, [self.forward.tree()])

    def get_dataset_names(self, names: Set[str]) -> None:
        """Get the dataset names.

        Parameters
        ----------
        names : Set[str]
            The set of dataset names.
        """
        super().get_dataset_names(names)
        self._statistic.get_dataset_names(names)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import itertools
import logging
import math
import time
from collections import defaultdict
from functools import cached_property
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
from anemoi.utils.dates import as_timedelta
from anemoi.utils.humanize import seconds_to_human
from anemoi.utils.humanize import shorten_list
from earthkit.data.core.order import build_remapping

from .action import ActionContext
from .trace import trace
from .trace import trace_datasource

LOG = logging.getLogger(__name__)


def _fields_metatata(variables: Tuple[str, ...], cube: Any) -> Dict[str, Any]:
    """Retrieve metadata for the given variables and cube.

    Parameters
    ----------
    variables : tuple of str
        The variables to retrieve metadata for.
    cube : Any
        The data cube.

    Returns
    -------
    dict
        The metadata dictionary.
    """
    assert isinstance(variables, tuple), variables

    KNOWN: Dict[str, Dict[str, bool]] = {
        "cos_julian_day": dict(computed_forcing=True, constant_in_time=False),
        "cos_latitude": dict(computed_forcing=True, constant_in_time=True),
        "cos_local_time": dict(computed_forcing=True, constant_in_time=False),
        "cos_longitude": dict(computed_forcing=True, constant_in_time=True),
        "cos_solar_zenith_angle": dict(computed_forcing=True, constant_in_time=False),
        "insolation": dict(computed_forcing=True, constant_in_time=False),
        "latitude": dict(computed_forcing=True, constant_in_time=True),
        "longitude": dict(computed_forcing=True, constant_in_time=True),
        "sin_julian_day": dict(computed_forcing=True, constant_in_time=False),
        "sin_latitude": dict(computed_forcing=True, constant_in_time=True),
        "sin_local_time": dict(computed_forcing=True, constant_in_time=False),
        "sin_longitude": dict(computed_forcing=True, constant_in_time=True),
    }

    def _merge(md1: Dict[str, Any], md2: Dict[str, Any]) -> Dict[str, Any]:
        assert set(md1.keys()) == set(md2.keys()), (set(md1.keys()), set(md2.keys()))
        result: Dict[str, Any] = {}
        for k in md1.keys():
            v1 = md1[k]
            v2 = md2[k]

            if v1 == v2:
                result[k] = v1
                continue

            if isinstance(v1, list):
                assert v2 not in v1, (v1, v2)
                result[k] = sorted(v1 + [v2])
                continue

            if isinstance(v2, list):
                assert v1 not in v2, (v1, v2)
                result[k] = sorted(v2 + [v1])
                continue

            result[k] = sorted([v1, v2])

        return result

    mars: Dict[str, Any] = {}
    other: DefaultDict[str, Dict[str, Any]] = defaultdict(dict)
    i: int = -1
    date: Optional[str] = None
    for c in cube.iterate_cubelets():

        if date is None:
            date = c._coords_names[0]

        if date != c._coords_names[0]:
            continue

        if i == -1 or c._coords_names[1] != variables[i]:
            i += 1

        f = cube[c.coords]
        md = f.metadata(namespace="mars")
        if not md:
            md = f.metadata(namespace="default")

        if md.get("param") == "~":
            md["param"] = f.metadata("param")
            assert md["param"] not in ("~", "unknown"), (md, f.metadata("param"))

        if md.get("param") == "unknown":
            md["param"] = str(f.metadata("paramId", default="unknown"))
            # assert md['param'] != 'unknown', (md, f.metadata('param'))

        startStep = f.metadata("startStep", default=None)
        if startStep is not None:
            startStep = as_timedelta(startStep)

        endStep = f.metadata("endStep", default=None)
        if endStep is not None:
            endStep = as_timedelta(endStep)

        stepTypeForConversion = f.metadata("stepTypeForConversion", default=None)
        typeOfStatisticalProcessing = f.metadata("typeOfStatisticalProcessing", default=None)
        timeRangeIndicator = f.metadata("timeRangeIndicator", default=None)

        # GRIB1 precipitation accumulations are not correctly encoded
        if startStep == endStep and stepTypeForConversion == "accum":
            startStep = 0

        if startStep != endStep:
            # https://codes.ecmwf.int/grib/format/grib2/ctables/4/10/
            TYPE_OF_STATISTICAL_PROCESSING: Dict[Optional[int], Optional[str]] = {
                None: None,
                0: "average",
                1: "accumulation",
                2: "maximum",
                3: "minimum",
                4: "difference(end-start)",
                5: "root_mean_square",
                6: "standard_deviation",
                7: "covariance",
                8: "difference(start-end)",
                9: "ratio",
                10: "standardized_anomaly",
                11: "summation",
                100: "severity",
                101: "mode",
            }

            # https://codes.ecmwf.int/grib/format/grib1/ctable/5/

            TIME_RANGE_INDICATOR: Dict[int, str] = {
                4: "accumulation",
                3: "average",
            }

            STEP_TYPE_FOR_CONVERSION: Dict[str, str] = {
                "min": "minimum",
                "max": "maximum",
                "accum": "accumulation",
            }

            #
            # A few patches
            #

            PATCHES: Dict[str, str] = {
                "10fg6": "maximum",
                "mntpr3": "minimum",  # Not in param db
                "mntpr6": "minimum",  # Not in param db
                "mxtpr3": "maximum",  # Not in param db
                "mxtpr6": "maximum",  # Not in param db
            }

            process = TYPE_OF_STATISTICAL_PROCESSING.get(typeOfStatisticalProcessing)
            if process is None:
                process = TIME_RANGE_INDICATOR.get(timeRangeIndicator)
            if process is None:
                process = STEP_TYPE_FOR_CONVERSION.get(stepTypeForConversion)
            if process is None:
                process = PATCHES.get(md["param"])
                if process is not None:
                    LOG.error(f"Unknown process {stepTypeForConversion} for {md['param']}, using {process} instead")

            if process is None:
                raise ValueError(
                    f"Unknown for {md['param']}:"
                    f" {stepTypeForConversion=} ({STEP_TYPE_FOR_CONVERSION.get('stepTypeForConversion')}),"
                    f" {typeOfStatisticalProcessing=} ({TYPE_OF_STATISTICAL_PROCESSING.get(typeOfStatisticalProcessing)}),"
                    f" {timeRangeIndicator=} ({TIME_RANGE_INDICATOR.get(timeRangeIndicator)})"
                )

            # print(md["param"], "startStep", startStep, "endStep", endStep, "process", process, "typeOfStatisticalProcessing", typeOfStatisticalProcessing)
            other[variables[i]]["process"] = process
            other[variables[i]]["period"] = (startStep, endStep)

        for k in md.copy().keys():
            if k.startswith("_"):
                md.pop(k)

        if variables[i] in mars:
            mars[variables[i]] = _merge(md, mars[variables[i]])
        else:
            mars[variables[i]] = md

    result: Dict[str, Dict[str, Any]] = {}
    for k, v in mars.items():
        result[k] = dict(mars=v) if v else {}
        result[k].update(other[k])
        result[k].update(KNOWN.get(k, {}))
        # assert result[k], k

    assert i + 1 == len(variables), (i + 1, len(variables))
    return result


def _data_request(data: Any) -> Dict[str, Any]:
    """Build a data request dictionary from the given data.

    Parameters
    ----------
    data : Any
        The data to build the request from.

    Returns
    -------
    dict
        The data request dictionary.
    """
    date: Optional[Any] = None
    params_levels: DefaultDict[str, set] = defaultdict(set)
    params_steps: DefaultDict[str, set] = defaultdict(set)

    area: Optional[Any] = None
    grid: Optional[Any] = None

    for field in data:
        try:
            if date is None:
                date = field.metadata("valid_datetime")

            if field.metadata("valid_datetime") != date:
                continue

            as_mars = field.metadata(namespace="mars")
            if not as_mars:
                continue
            step = as_mars.get("step")
            levtype = as_mars.get("levtype", "sfc")
            param = as_mars["param"]
            levelist = as_mars.get("levelist", None)
            area = field.mars_area
            grid = field.mars_grid

            if levelist is None:
                params_levels[levtype].add(param)
            else:
                params_levels[levtype].add((param, levelist))

            if step:
                params_steps[levtype].add((param, step))
        except Exception:
            LOG.error(f"Error in retrieving metadata (cannot build data request info) for {field}", exc_info=True)

    def sort(old_dic: DefaultDict[str, set]) -> Dict[str, List[Any]]:
        new_dic: Dict[str, List[Any]] = {}
        for k, v in old_dic.items():
            new_dic[k] = sorted(list(v))
        return new_dic

    params_steps = sort(params_steps)
    params_levels = sort(params_levels)

    return dict(param_level=params_levels, param_step=params_steps, area=area, grid=grid)


class Result:
    """Class to represent the result of an action in the dataset creation process."""

    empty: bool = False
    _coords_already_built: bool = False

    def __init__(self, context: ActionContext, action_path: List[str], dates: Any) -> None:
        """Initialize a Result instance.

        Parameters
        ----------
        context : ActionContext
            The context in which the result exists.
        action_path : list of str
            The action path.
        dates : Any
            The dates associated with the result.
        """
        from anemoi.datasets.dates.groups import GroupOfDates

        assert isinstance(dates, GroupOfDates), dates

        assert isinstance(context, ActionContext), type(context)
        assert isinstance(action_path, list), action_path

        self.context: Any = context
        self.group_of_dates: Any = dates
        self.action_path: List[str] = action_path

    @property
    @trace_datasource
    def datasource(self) -> Any:
        """Retrieve the data source for the result."""
        self._raise_not_implemented()

    @property
    def data_request(self) -> Dict[str, Any]:
        """Returns a dictionary with the parameters needed to retrieve the data."""
        return _data_request(self.datasource)

    def get_cube(self) -> Any:
        """Retrieve the data cube for the result.

        Returns
        -------
        Any
            The data cube.
        """
        trace("🧊", f"getting cube from {self.__class__.__name__}")
        ds: Any = self.datasource

        remapping: Any = self.context.remapping
        order_by: Any = self.context.order_by
        flatten_grid: Any = self.context.flatten_grid
        start: float = time.time()
        LOG.debug("Sorting dataset %s %s", dict(order_by), remapping)
        assert order_by, order_by

        patches: Dict[str, Dict[Optional[Any], int]] = {"number": {None: 0}}

        try:
            cube: Any = ds.cube(
                order_by,
                remapping=remapping,
                flatten_values=flatten_grid,
                patches=patches,
            )
            cube = cube.squeeze()
            LOG.debug(f"Sorting done in {seconds_to_human(time.time()-start)}.")
        except ValueError:
            self.explain(ds, order_by, remapping=remapping, patches=patches)
            # raise ValueError(f"Error in {self}")
            exit(1)

        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug("Cube shape: %s", cube)
            for k, v in cube.user_coords.items():
                LOG.debug("  %s %s", k, shorten_list(v, max_length=10))

        return cube

    def explain(self, ds: Any, *args: Any, remapping: Any, patches: Any) -> None:
        """Explain the data cube creation process.

        Parameters
        ----------
        ds : Any
            The data source.
        args : Any
            Additional arguments.
        remapping : Any
            The remapping configuration.
        patches : Any
            The patches configuration.
        """
        METADATA: Tuple[str, ...] = (
            "date",
            "time",
            "step",
            "hdate",
            "valid_datetime",
            "levtype",
            "levelist",
            "number",
            "level",
            "shortName",
            "paramId",
            "variable",
        )

        # We redo the logic here
        print()
        print("❌" * 40)
        print()
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]

        # print("Executing", self.action_path)
        # print("Dates:", compress_dates(self.dates))

        names: List[str] = []
        for a in args:
            if isinstance(a, str):
                names.append(a)
            elif isinstance(a, dict):
                names += list(a.keys())

        print(f"Building a {len(names)}D hypercube using", names)
        ds = ds.order_by(*args, remapping=remapping, patches=patches)
        user_coords = ds.unique_values(*names, remapping=remapping, patches=patches, progress_bar=False)

        print()
        print("Number of unique values found for each coordinate:")
        for k, v in user_coords.items():
            print(f"  {k:20}:", len(v), shorten_list(v, max_length=10))
        print()
        user_shape: Tuple[int, ...] = tuple(len(v) for k, v in user_coords.items())
        print("Shape of the hypercube           :", user_shape)
        print(
            "Number of expected fields        :", math.prod(user_shape), "=", " x ".join([str(i) for i in user_shape])
        )
        print("Number of fields in the dataset  :", len(ds))
        print("Difference                       :", abs(len(ds) - math.prod(user_shape)))
        print()

        remapping = build_remapping(remapping, patches)
        expected = set(itertools.product(*user_coords.values()))
        extra = set()

        if math.prod(user_shape) > len(ds):
            print(f"This means that all the fields in the datasets do not exists for all combinations of {names}.")

            for f in ds:
                metadata = remapping(f.metadata)
                key = tuple(metadata(n, default=None) for n in names)
                if key in expected:
                    expected.remove(key)
                else:
                    extra.add(key)

            print("Missing fields:")
            print()
            for i, f in enumerate(sorted(expected)):
                print(" ", f)
                if i >= 9 and len(expected) > 10:
                    print("...", len(expected) - i - 1, "more")
                    break

            print("Extra fields:")
            print()
            for i, f in enumerate(sorted(extra)):
                print(" ", f)
                if i >= 9 and len(extra) > 10:
                    print("...", len(extra) - i - 1, "more")
                    break

            print()
            print("Missing values:")
            per_name = defaultdict(set)
            for e in expected:
                for n, v in zip(names, e):
                    per_name[n].add(v)

            for n, v in per_name.items():
                print(" ", n, len(v), shorten_list(sorted(v), max_length=10))
            print()

            print("Extra values:")
            per_name = defaultdict(set)
            for e in extra:
                for n, v in zip(names, e):
                    per_name[n].add(v)

            for n, v in per_name.items():
                print(" ", n, len(v), shorten_list(sorted(v), max_length=10))
            print()

            print("To solve this issue, you can:")
            print(
                "  - Provide a better selection, like 'step: 0' or 'level: 1000' to "
                "reduce the number of selected fields."
            )
            print(
                "  - Split the 'input' part in smaller sections using 'join', "
                "making sure that each section represent a full hypercube."
            )

        else:
            print(f"More fields in dataset that expected for {names}. " "This means that some fields are duplicated.")
            duplicated = defaultdict(list)
            for f in ds:
                # print(f.metadata(namespace="default"))
                metadata = remapping(f.metadata)
                key = tuple(metadata(n, default=None) for n in names)
                duplicated[key].append(f)

            print("Duplicated fields:")
            print()
            duplicated = {k: v for k, v in duplicated.items() if len(v) > 1}
            for i, (k, v) in enumerate(sorted(duplicated.items())):
                print(" ", k)
                for f in v:
                    x = {k: f.metadata(k, default=None) for k in METADATA if f.metadata(k, default=None) is not None}
                    print("   ", f, x)
                if i >= 9 and len(duplicated) > 10:
                    print("...", len(duplicated) - i - 1, "more")
                    break

            print()
            print("To solve this issue, you can:")
            print("  - Provide a better selection, like 'step: 0' or 'level: 1000'")
            print("  - Change the way 'param' is computed using 'variable_naming' " "in the 'build' section.")

        print()
        print("❌" * 40)
        print()
        exit(1)

    def _repr(self, *args: Any, _indent_: str = "\n", **kwargs: Any) -> str:
        """Return the string representation of the Result instance.

        Parameters
        ----------
        args : Any
            Additional positional arguments.
        _indent_ : str
            Indentation string.
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        str
            The string representation.
        """
        more: str = ",".join([str(a)[:5000] for a in args])
        more += ",".join([f"{k}={v}"[:5000] for k, v in kwargs.items()])

        dates: str = " no-dates"
        if self.group_of_dates is not None:
            dates = f" {len(self.group_of_dates)} dates"
            dates += " ("
            dates += "/".join(d.strftime("%Y-%m-%dT%H:%M") for d in self.group_of_dates)
            if len(dates) > 100:
                dates = dates[:100] + "..."
            dates += ")"

        more = more[:5000]
        txt: str = f"{self.__class__.__name__}:{dates}{_indent_}{more}"
        if _indent_:
            txt = txt.replace("\n", "\n  ")
        return txt

    def __repr__(self) -> str:
        """Return the string representation of the Result instance."""
        return self._repr()

    def _raise_not_implemented(self) -> None:
        """Raise a NotImplementedError indicating the method is not implemented."""
        raise NotImplementedError(f"Not implemented in {self.__class__.__name__}")

    def _trace_datasource(self, *args: Any, **kwargs: Any) -> str:
        """Trace the data source for the result.

        Parameters
        ----------
        args : Any
            Additional positional arguments.
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        str
            The trace string.
        """
        return f"{self.__class__.__name__}({self.group_of_dates})"

    def build_coords(self) -> None:
        """Build the coordinates for the result."""
        if self._coords_already_built:
            return

        cube: Any = self.get_cube()

        from_data: Any = cube.user_coords
        from_config: Any = self.context.order_by

        keys_from_config: list = list(from_config.keys())
        keys_from_data: list = list(from_data.keys())
        assert keys_from_data == keys_from_config, f"Critical error: {keys_from_data=} != {keys_from_config=}. {self=}"

        variables_key: str = list(from_config.keys())[1]
        ensembles_key: str = list(from_config.keys())[2]

        if isinstance(from_config[variables_key], (list, tuple)):
            assert all([v == w for v, w in zip(from_data[variables_key], from_config[variables_key])]), (
                from_data[variables_key],
                from_config[variables_key],
            )

        self._variables: Any = from_data[variables_key]  # "param_level"
        self._ensembles: Any = from_data[ensembles_key]  # "number"

        first_field: Any = self.datasource[0]
        grid_points: Any = first_field.grid_points()

        lats: Any = grid_points[0]
        lons: Any = grid_points[1]

        assert len(lats) == len(lons), (len(lats), len(lons), first_field)
        assert len(lats) == math.prod(first_field.shape), (len(lats), first_field.shape, first_field)

        north: float = np.amax(lats)
        south: float = np.amin(lats)
        east: float = np.amax(lons)
        west: float = np.amin(lons)

        assert -90 <= south <= north <= 90, (south, north, first_field)
        assert (-180 <= west <= east <= 180) or (0 <= west <= east <= 360), (
            west,
            east,
            first_field,
        )

        grid_values: list = list(range(len(grid_points[0])))

        self._grid_points: Any = grid_points
        self._resolution: Any = first_field.resolution
        self._grid_values: Any = grid_values
        self._field_shape: Any = first_field.shape
        self._proj_string: Any = first_field.proj_string if hasattr(first_field, "proj_string") else None

        self._cube: Any = cube

        self._coords_already_built: bool = True

    @property
    def variables(self) -> List[str]:
        """Retrieve the variables for the result."""
        self.build_coords()
        return self._variables

    @property
    def variables_metadata(self) -> Dict[str, Any]:
        """Retrieve the metadata for the variables."""
        return _fields_metatata(self.variables, self._cube)

    @property
    def ensembles(self) -> Any:
        """Retrieve the ensembles for the result."""
        self.build_coords()
        return self._ensembles

    @property
    def resolution(self) -> Any:
        """Retrieve the resolution for the result."""
        self.build_coords()
        return self._resolution

    @property
    def grid_values(self) -> Any:
        """Retrieve the grid values for the result."""
        self.build_coords()
        return self._grid_values

    @property
    def grid_points(self) -> Any:
        """Retrieve the grid points for the result."""
        self.build_coords()
        return self._grid_points

    @property
    def field_shape(self) -> Any:
        """Retrieve the field shape for the result."""
        self.build_coords()
        return self._field_shape

    @property
    def proj_string(self) -> Any:
        """Retrieve the projection string for the result."""
        self.build_coords()
        return self._proj_string

    @cached_property
    def shape(self) -> List[int]:
        """Retrieve the shape of the result."""
        return [
            len(self.group_of_dates),
            len(self.variables),
            len(self.ensembles),
            len(self.grid_values),
        ]

    @cached_property
    def coords(self) -> Dict[str, Any]:
        """Retrieve the coordinates of the result."""
        return {
            "dates": list(self.group_of_dates),
            "variables": self.variables,
            "ensembles": self.ensembles,
            "values": self.grid_values,
        }



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import glob
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import earthkit.data as ekd
from anemoi.transform.flavour import RuleBasedFlavour
from earthkit.data import from_source
from earthkit.data.indexing.fieldlist import FieldArray
from earthkit.data.utils.patterns import Pattern

from .legacy import legacy_source

LOG = logging.getLogger(__name__)


def _load(context: Any, name: str, record: Dict[str, Any]) -> tuple:
    """Load data from a given source.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    name : str
        The name of the data source.
    record : dict of str to Any
        The record containing source information.

    Returns
    -------
    tuple
        A tuple containing the data as a numpy array and the UUID of the HGrid.
    """
    ds = None

    param = record["param"]

    if "path" in record:
        context.info(f"Using {name} from {record['path']} (param={param})")
        ds = from_source("file", record["path"])

    if "url" in record:
        context.info(f"Using {name} from {record['url']} (param={param})")
        ds = from_source("url", record["url"])

    ds = ds.sel(param=param)

    assert len(ds) == 1, f"{name} {param}, expected one field, got {len(ds)}"
    ds = ds[0]

    return ds.to_numpy(flatten=True), ds.metadata("uuidOfHGrid")


class Geography:
    """This class retrieves the latitudes and longitudes of unstructured grids,
    and checks if the fields are compatible with the grid.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    latitudes : dict of str to Any
        Latitude information.
    longitudes : dict of str to Any
        Longitude information.
    """

    def __init__(self, context: Any, latitudes: Dict[str, Any], longitudes: Dict[str, Any]) -> None:
        """Initialize the Geography class.

        Parameters
        ----------
        context : Any
            The context in which the function is executed.
        latitudes : dict of str to Any
            Latitude information.
        longitudes : dict of str to Any
            Longitude information.
        """
        latitudes, uuidOfHGrid_lat = _load(context, "latitudes", latitudes)
        longitudes, uuidOfHGrid_lon = _load(context, "longitudes", longitudes)

        assert (
            uuidOfHGrid_lat == uuidOfHGrid_lon
        ), f"uuidOfHGrid mismatch: lat={uuidOfHGrid_lat} != lon={uuidOfHGrid_lon}"

        context.info(f"Latitudes: {len(latitudes)}, Longitudes: {len(longitudes)}")
        assert len(latitudes) == len(longitudes)

        self.uuidOfHGrid = uuidOfHGrid_lat
        self.latitudes = latitudes
        self.longitudes = longitudes
        self.first = True

    def check(self, field: Any) -> None:
        """Check if the field is compatible with the grid.

        Parameters
        ----------
        field : Any
            The field to check.
        """
        if self.first:
            # We only check the first field, for performance reasons
            assert (
                field.metadata("uuidOfHGrid") == self.uuidOfHGrid
            ), f"uuidOfHGrid mismatch: {field.metadata('uuidOfHGrid')} != {self.uuidOfHGrid}"
            self.first = False


class AddGrid:
    """An earth-kit.data.Field wrapper that adds grid information.

    Parameters
    ----------
    field : Any
        The field to wrap.
    geography : Geography
        The geography information.
    """

    def __init__(self, field: Any, geography: Geography) -> None:
        """Initialize the AddGrid class.

        Parameters
        ----------
        field : Any
            The field to wrap.
        geography : Geography
            The geography information.
        """
        self._field = field

        geography.check(field)

        self._latitudes = geography.latitudes
        self._longitudes = geography.longitudes

    def __getattr__(self, name: str) -> Any:
        """Get an attribute from the wrapped field.

        Parameters
        ----------
        name : str
            The name of the attribute.

        Returns
        -------
        Any
            The attribute value.
        """
        return getattr(self._field, name)

    def __repr__(self) -> str:
        """Get the string representation of the wrapped field.

        Returns
        -------
        str
            The string representation.
        """
        return repr(self._field)

    def grid_points(self) -> tuple:
        """Get the grid points (latitudes and longitudes).

        Returns
        -------
        tuple
            The latitudes and longitudes.
        """
        return self._latitudes, self._longitudes

    @property
    def resolution(self) -> str:
        """Get the resolution of the grid."""
        return "unknown"


def check(ds: Any, paths: List[str], **kwargs: Any) -> None:
    """Check if the dataset matches the expected number of fields.

    Parameters
    ----------
    ds : Any
        The dataset to check.
    paths : list of str
        List of paths to the GRIB files.
    **kwargs : Any
        Additional keyword arguments.

    Raises
    ------
    ValueError
        If the number of fields does not match the expected count.
    """
    count = 1
    for k, v in kwargs.items():
        if isinstance(v, (tuple, list)):
            count *= len(v)

    if len(ds) != count:
        raise ValueError(f"Expected {count} fields, got {len(ds)} (kwargs={kwargs}, paths={paths})")


def _expand(paths: List[str]) -> Any:
    """Expand the given paths using glob.

    Parameters
    ----------
    paths : list of str
        List of paths to expand.

    Returns
    -------
    Any
        The expanded paths.
    """
    for path in paths:
        cnt = 0
        for p in glob.glob(path):
            yield p
            cnt += 1
        if cnt == 0:
            yield path


@legacy_source(__file__)
def execute(
    context: Any,
    dates: List[Any],
    path: Union[str, List[str]],
    latitudes: Optional[Dict[str, Any]] = None,
    longitudes: Optional[Dict[str, Any]] = None,
    flavour: Optional[Union[str, Dict[str, Any]]] = None,
    *args: Any,
    **kwargs: Any,
) -> ekd.FieldList:
    """Execute the function to load data from GRIB files.

    Args:
        context (Any): The context in which the function is executed.
        dates (List[Any]): List of dates.
        path (Union[str, List[str]]): Path or list of paths to the GRIB files.
        latitudes (Optional[Dict[str, Any]], optional): Latitude information. Defaults to None.
        longitudes (Optional[Dict[str, Any]], optional): Longitude information. Defaults to None.
        flavour (Optional[Union[str, Dict[str, Any]]], optional): Flavour information. Defaults to None.
        *args (Any): Additional arguments.
        **kwargs (Any): Additional keyword arguments.

    Returns
    -------
    Any
        The loaded dataset.
    """
    given_paths = path if isinstance(path, list) else [path]
    if flavour is not None:
        flavour = RuleBasedFlavour(flavour)

    geography = None
    if latitudes is not None and longitudes is not None:
        geography = Geography(context, latitudes, longitudes)

    ds = from_source("empty")
    dates = [d.isoformat() for d in dates]

    for path in given_paths:
        paths = Pattern(path, ignore_missing_keys=True).substitute(*args, date=dates, **kwargs)

        for name in ("grid", "area", "rotation", "frame", "resol", "bitmap"):
            if name in kwargs:
                raise ValueError(f"MARS interpolation parameter '{name}' not supported")

        for path in _expand(paths):
            context.trace("📁", "PATH", path)
            s = from_source("file", path)
            if flavour is not None:
                s = flavour.map(s)
            s = s.sel(valid_datetime=dates, **kwargs)
            ds = ds + s

    if kwargs and not context.partial_ok:
        check(ds, given_paths, valid_datetime=dates, **kwargs)

    if geography is not None:
        ds = FieldArray([AddGrid(_, geography) for _ in ds])

    if len(ds) == 0:
        LOG.warning(f"No fields found for {dates} in {given_paths} (kwargs={kwargs})")

    return ds



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from collections import defaultdict
from typing import Any
from typing import Dict
from typing import Tuple

import earthkit.data as ekd
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list
from earthkit.meteo import thermo

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(context: Any, input: ekd.FieldList, t: str, rh: str, q: str = "q") -> ekd.FieldList:
    """Convert relative humidity on pressure levels to specific humidity.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    input : List[Any]
        List of input fields.
    t : str
        Temperature parameter.
    rh : str
        Relative humidity parameter.
    q : str, optional
        Specific humidity parameter. Defaults to "q".

    Returns
    -------
    ekd.FieldList
        Array of fields with specific humidity.
    """
    result = []
    params: Tuple[str, str] = (t, rh)
    pairs: Dict[Tuple[Any, ...], Dict[str, Any]] = defaultdict(dict)

    # Gather all necessary fields
    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")
        if param in params:
            key = tuple(key.items())

            if param in pairs[key]:
                raise ValueError(f"Duplicate field {param} for {key}")

            pairs[key][param] = f
            if param == t:
                result.append(f)
        # all other parameters
        else:
            result.append(f)

    for keys, values in pairs.items():
        # some checks

        if len(values) != 2:
            raise ValueError("Missing fields")

        t_pl = values[t].to_numpy(flatten=True)
        rh_pl = values[rh].to_numpy(flatten=True)
        pressure = next(
            float(v) * 100 for k, v in keys if k in ["level", "levelist"]
        )  # Looks first for "level" then "levelist" value
        # print(f"Handling fields for pressure level {pressure}...")

        # actual conversion from rh --> q_v
        q_pl = thermo.specific_humidity_from_relative_humidity(t_pl, rh_pl, pressure)
        result.append(new_field_from_numpy(values[rh], q_pl, param=q))

    return new_fieldlist_from_list(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import math
import os

from anemoi.datasets import open_dataset

from . import Command

RADIUS_EARTH_KM = 6371.0  # Earth's radius in kilometers

LOG = logging.getLogger(__name__)


class HTML_Writer:
    def __init__(self):
        self.html_content = """
        <!DOCTYPE html>
        <html>
        <head>
            <style>
                table {
                    border-collapse: collapse;
                    width: 100%;
                }
                th, td {
                    border: 1px solid black;
                    padding: 8px;
                    text-align: center;
                }
                th {
                    background-color: #e2e2e2;
                }
            </style>
        </head>
        <body>
            <table>
                <thead>
                    <tr>
                        <th>Variable</th>
                        <th>Global Mean</th>
                        <th>LAM Mean</th>
                        <th>Mean Diff (%)</th>
                        <th>Global Std</th>
                        <th>LAM Std</th>
                        <th>Std Diff (%)</th>
                        <th>Global Max</th>
                        <th>LAM Max</th>
                        <th>Max Diff (%)</th>
                        <th>Global Min</th>
                        <th>LAM Min</th>
                        <th>Min Diff (%)</th>
                    </tr>
                </thead>
                <tbody>
        """

    def update_table(
        self,
        v1,
        global_mean,
        lam_mean,
        mean_diff,
        global_std,
        lam_std,
        std_diff,
        global_max,
        lam_max,
        max_diff,
        global_min,
        lam_min,
        min_diff,
    ):

        # Determine inline style for HTML
        mean_bg_color = "background-color: #d4edda;" if abs(mean_diff) < 20 else "background-color: #f8d7da;"
        std_bg_color = "background-color: #d4edda;" if abs(std_diff) < 20 else "background-color: #f8d7da;"
        max_bg_color = "background-color: #d4edda;" if abs(max_diff) < 20 else "background-color: #f8d7da;"
        min_bg_color = "background-color: #d4edda;" if abs(min_diff) < 20 else "background-color: #f8d7da;"

        # Add a row to the HTML table with inline styles
        self.html_content += f"""
            <tr>
                <td style="background-color: #f2f2f2;">{v1}</td>
                <td>{global_mean}</td>
                <td>{lam_mean}</td>
                <td style="{mean_bg_color}">{mean_diff}%</td>
                <td>{global_std}</td>
                <td>{lam_std}</td>
                <td style="{std_bg_color}">{std_diff}%</td>
                <td>{global_max}</td>
                <td>{lam_max}</td>
                <td style="{max_bg_color}">{max_diff}%</td>
                <td>{global_min}</td>
                <td>{lam_min}</td>
                <td style="{min_bg_color}">{min_diff}%</td>
            </tr>
        """

    def save_table(self, save_path="stats_table.html"):
        # Close the HTML tags
        self.html_content += """
                </tbody>
            </table>
        </body>
        </html>
        """

        # Save the HTML content to a file
        with open(save_path, "w") as f:
            f.write(self.html_content)

        LOG.info(f"\nHTML table saved to: {save_path}")


def plot_coordinates_on_map(lats, lons):
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature
    import matplotlib.pyplot as plt

    """
    Plots the given latitude and longitude coordinates on a map using Cartopy and Matplotlib.

    Parameters:
    - lats: List of latitudes
    - lons: List of longitudes
    """

    if len(lats) != len(lons):
        raise ValueError("The length of latitude and longitude lists must be the same.")

    # Create a figure and axis using the PlateCarree projection

    # Define source (PlateCarree) and target (LambertConformal) projections
    target_proj = ccrs.LambertConformal(central_latitude=0, central_longitude=10, standard_parallels=[63.3, 63.3])

    # Create a figure and axis
    fig, ax = plt.subplots(figsize=(14, 12), subplot_kw={"projection": target_proj})

    # Set the extent of the map based on the transformed coordinates
    margin = 10
    ax.set_extent(
        [min(lons) - margin, max(lons) + margin, min(lats) - margin, max(lats) + margin], crs=ccrs.PlateCarree()
    )
    # ax.set_extent([-25, 45, 30, 75], crs=ccrs.PlateCarree())

    # Add map features
    ax.add_feature(cfeature.LAND)
    ax.add_feature(cfeature.OCEAN)
    ax.add_feature(cfeature.COASTLINE.with_scale("50m"), zorder=1, alpha=0.8)
    ax.add_feature(cfeature.BORDERS.with_scale("50m"), linestyle=":", zorder=1)

    # Plot transformed coordinates
    ax.scatter(lons, lats, color="blue", s=1, edgecolor="b", transform=ccrs.PlateCarree(), alpha=0.3)
    ax.set_title("Latitude and Longitude")
    ax.title.set_size(20)

    # Show the plot
    return fig


def haversine(lat1, lon1, lat2, lon2):
    """Calculate the great-circle distance between two points on the Earth's surface using the Haversine formula."""
    dlat = math.radians(lat2 - lat1)
    dlon = math.radians(lon2 - lon1)
    a = math.sin(dlat / 2) ** 2 + math.cos(math.radians(lat1)) * math.cos(math.radians(lat2)) * math.sin(dlon / 2) ** 2
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    return RADIUS_EARTH_KM * c


def rectangle_area_km2(lat1, lon1, lat2, lon2):
    """Calculate the area of a rectangle given the coordinates of the top left and bottom right corners in
    latitude and longitude.

    Parameters:
    lat1, lon1 - Latitude and longitude of the top-left corner.
    lat2, lon2 - Latitude and longitude of the bottom-right corner.

    Returns:
    Area in square kilometers (km^2).
    """
    # Calculate the height (difference in latitude)
    height_km = haversine(lat1, lon1, lat2, lon1)

    # Calculate the width (difference in longitude)
    width_km = haversine(lat1, lon1, lat1, lon2)

    # Area of the rectangle
    area_km2 = height_km * width_km
    return area_km2


def check_order(vars_1, vars_2):
    for v1, v2 in zip(vars_1, vars_2):
        if v1 != v2 and ((v1 in vars_2) and (v2 in vars_1)):
            return False

    return True


def compute_wighted_diff(s1, s2, round_ndigits):
    return round((s2 - s1) * 100 / s2, ndigits=round_ndigits)


class CompareLAM(Command):
    """Compare statistic of two datasets. \
       This command compares the statistics of each variable in two datasets ONLY in the overlapping area between the two. \
    """

    def add_arguments(self, command_parser):
        command_parser.add_argument("dataset1", help="Path of the global dataset or the largest dataset.")
        command_parser.add_argument("dataset2", help="Path of the LAM dataset or the smallest dataset.")
        command_parser.add_argument(
            "-D", "--number-of-dates", type=int, default=10, help="Number of datapoints (in time) to compare over."
        )
        command_parser.add_argument("-O", "--outpath", default="./", help="Path to output folder.")
        command_parser.add_argument("-R", "--number-of-digits", type=int, default=4, help="Number of digits to keep.")
        command_parser.add_argument(
            "--selected-vars",
            nargs="+",
            default=["10u", "10v", "2d", "2t"],
            help="List of selected variables to use in the script.",
        )
        command_parser.add_argument(
            "--save-plots", action="store_true", help="Toggle to save a picture of the data grid."
        )

    def run(self, args):
        import matplotlib.pyplot as plt
        import numpy as np
        from prettytable import PrettyTable
        from termcolor import colored  # For coloring text in the terminal

        # Unpack args
        date_idx = args.number_of_dates
        round_ndigits = args.number_of_digits
        selected_vars = args.selected_vars
        global_name = args.dataset1
        lam_name = args.dataset2
        date_idx = 10  # "all" or specific index to stop at
        name = f"{global_name}-{lam_name}_{date_idx}"
        save_path = os.path.join(args.outpath, f"comparison_table_{name}.html")

        # Open LAM dataset
        lam_dataset = open_dataset(lam_name, select=selected_vars)
        lam_vars = list(lam_dataset.variables)
        lam_num_grid_points = lam_dataset[0, 0].shape[1]
        lam_area = rectangle_area_km2(
            max(lam_dataset.latitudes),
            max(lam_dataset.longitudes),
            min(lam_dataset.latitudes),
            min(lam_dataset.longitudes),
        )
        l_coords = (
            max(lam_dataset.latitudes),
            min(lam_dataset.longitudes),
            min(lam_dataset.latitudes),
            max(lam_dataset.longitudes),
        )

        if args.save_plots:
            _ = plot_coordinates_on_map(lam_dataset.latitudes, lam_dataset.longitudes)
            plt.savefig(os.path.join(args.outpath, "lam_dataset.png"))

        LOG.info(f"Dataset {lam_name}, has {lam_num_grid_points} grid points. \n")
        LOG.info("LAM (north, west, south, east): ", l_coords)
        LOG.info(f"Point every: {math.sqrt(lam_area / lam_num_grid_points)} km")

        # Open global dataset and cut it
        lam_start = lam_dataset.dates[0]
        lam_end = lam_dataset.dates[-1]
        global_dataset = open_dataset(global_name, start=lam_start, end=lam_end, area=l_coords, select=selected_vars)
        global_vars = list(global_dataset.variables)
        global_num_grid_points = global_dataset[0, 0].shape[1]
        global_area = rectangle_area_km2(
            max(global_dataset.latitudes),
            max(global_dataset.longitudes),
            min(global_dataset.latitudes),
            min(global_dataset.longitudes),
        )
        g_coords = (
            max(global_dataset.latitudes),
            min(global_dataset.longitudes),
            min(global_dataset.latitudes),
            max(global_dataset.longitudes),
        )

        if args.save_plots:
            _ = plot_coordinates_on_map(global_dataset.latitudes, global_dataset.longitudes)
            plt.savefig(os.path.join(args.outpath, "global_dataset.png"))

        LOG.info(f"Dataset {global_name}, has {global_num_grid_points} grid points. \n")
        LOG.info("Global-lam cut (north, west, south, east): ", g_coords)
        LOG.info(f"Point every: {math.sqrt(global_area / global_num_grid_points)} km")

        # Check variable ordering
        same_order = check_order(global_vars, lam_vars)
        LOG.info(f"Lam dataset has the same order of variables as the global dataset: {same_order}")

        LOG.info("\nComparing statistics..")
        table = PrettyTable()
        table.field_names = [
            "Variable",
            "Global Mean",
            "LAM Mean",
            "Mean Diff (%)",
            "Global Std",
            "LAM Std",
            "Std Diff (%)",
            "Global Max",
            "LAM Max",
            "Max Diff (%)",
            "Global Min",
            "LAM Min",
            "Min Diff (%)",
        ]

        # Create a styled HTML table
        html_writer = HTML_Writer()

        for v1, v2 in zip(global_vars, lam_vars):
            assert v1 == v2
            idx = global_vars.index(v1)

            if date_idx == "all":
                lam_mean = lam_dataset.statistics["mean"][idx]
                lam_std = lam_dataset.statistics["stdev"][idx]
                lam_max = lam_dataset.statistics["max"][idx]
                lam_min = lam_dataset.statistics["min"][idx]
                global_mean = global_dataset.statistics["mean"][idx]
                global_std = global_dataset.statistics["stdev"][idx]
                global_max = global_dataset.statistics["max"][idx]
                global_min = global_dataset.statistics["min"][idx]

            else:
                lam_mean = np.nanmean(lam_dataset[:date_idx], axis=(0, 3))[idx][0]
                lam_std = np.nanstd(lam_dataset[:date_idx], axis=(0, 3))[idx][0]
                lam_max = np.nanmax(lam_dataset[:date_idx], axis=(0, 3))[idx][0]
                lam_min = np.nanmin(lam_dataset[:date_idx], axis=(0, 3))[idx][0]
                global_mean = np.nanmean(global_dataset[:date_idx], axis=(0, 3))[idx][0]
                global_std = np.nanstd(global_dataset[:date_idx], axis=(0, 3))[idx][0]
                global_max = np.nanmax(global_dataset[:date_idx], axis=(0, 3))[idx][0]
                global_min = np.nanmin(global_dataset[:date_idx], axis=(0, 3))[idx][0]

            mean_diff = compute_wighted_diff(lam_mean, global_mean, round_ndigits)
            std_diff = compute_wighted_diff(lam_std, global_std, round_ndigits)
            max_diff = compute_wighted_diff(lam_max, global_max, round_ndigits)
            min_diff = compute_wighted_diff(lam_min, global_min, round_ndigits)

            mean_color = "red" if abs(mean_diff) >= 20 else "green"
            std_color = "red" if abs(std_diff) >= 20 else "green"
            max_color = "red" if abs(max_diff) >= 20 else "green"
            min_color = "red" if abs(min_diff) >= 20 else "green"

            table.add_row(
                [
                    v1,
                    round(global_mean, ndigits=round_ndigits),
                    round(lam_mean, ndigits=round_ndigits),
                    colored(f"{mean_diff}%", mean_color),
                    round(global_std, ndigits=round_ndigits),
                    round(lam_std, ndigits=round_ndigits),
                    colored(f"{std_diff}%", std_color),
                    round(global_max, ndigits=round_ndigits),
                    round(lam_max, ndigits=round_ndigits),
                    colored(f"{max_diff}%", max_color),
                    round(global_min, ndigits=round_ndigits),
                    round(lam_min, ndigits=round_ndigits),
                    colored(f"{min_diff}%", min_color),
                ]
            )

            html_writer.update_table(
                v1,
                global_mean,
                lam_mean,
                mean_diff,
                global_std,
                lam_std,
                std_diff,
                global_max,
                lam_max,
                max_diff,
                global_min,
                lam_min,
                min_diff,
            )

        html_writer.save_table(save_path)
        print(table)


command = CompareLAM



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from numpy.typing import NDArray

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import Node
from .debug import Source
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class Select(Forwards):
    """Class to select a subset of variables from a dataset."""

    def __init__(self, dataset: Dataset, indices: List[int], reason: Dict[str, Any]) -> None:
        """Initialize the Select class.

        Parameters
        ----------
        dataset : Dataset
            The dataset to select from.
        indices : List[int]
            The indices of the variables to select.
        reason : Dict[str, Any]
            The reason for the selection.
        """
        reason = reason.copy()

        while isinstance(dataset, Select):
            indices = [dataset.indices[i] for i in indices]
            reason.update(dataset.reason)
            dataset = dataset.dataset

        self.dataset: Dataset = dataset
        self.indices = list(indices)
        assert len(self.indices) > 0
        self.reason = reason or {"indices": self.indices}

        super().__init__(dataset)

    def clone(self, dataset: Dataset) -> Dataset:
        """Clone the Select object with a new dataset.

        Parameters
        ----------
        dataset : Dataset
            The new dataset.

        Returns
        -------
        Select
            The cloned Select object.
        """
        return self.__class__(dataset, self.indices, self.reason).mutate()

    def mutate(self) -> Dataset:
        """Mutate the dataset.

        Returns
        -------
        Dataset
            The mutated dataset.
        """
        return self.forward.swap_with_parent(parent=self)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Get a tuple of data.

        Parameters
        ----------
        index : TupleIndex
            The index to retrieve.

        Returns
        -------
        NDArray[Any]
            The retrieved data.
        """
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, 1, slice(None))
        result = self.dataset[index]
        result = result[:, self.indices]
        result = result[:, previous]
        result = apply_index_to_slices_changes(result, changes)
        return result

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Get an item from the dataset.

        Parameters
        ----------
        n : FullIndex
            The index to retrieve.

        Returns
        -------
        NDArray[Any]
            The retrieved data.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        row = self.dataset[n]
        if isinstance(n, slice):
            return row[:, self.indices]

        return row[self.indices]

    @cached_property
    def shape(self) -> Shape:
        """Get the shape of the dataset."""
        return (len(self), len(self.indices)) + self.dataset.shape[2:]

    @cached_property
    def variables(self) -> List[str]:
        """Get the variables of the dataset."""
        return [self.dataset.variables[i] for i in self.indices]

    @cached_property
    def variables_metadata(self) -> Dict[str, Any]:
        """Get the metadata of the variables."""
        return {k: v for k, v in self.dataset.variables_metadata.items() if k in self.variables}

    @cached_property
    def name_to_index(self) -> Dict[str, int]:
        """Get the mapping of variable names to indices."""
        return {k: i for i, k in enumerate(self.variables)}

    @cached_property
    def statistics(self) -> Dict[str, NDArray[Any]]:
        """Get the statistics of the dataset."""
        return {k: v[self.indices] for k, v in self.dataset.statistics.items()}

    def statistics_tendencies(self, delta: Optional[datetime.timedelta] = None) -> Dict[str, NDArray[Any]]:
        """Get the statistical tendencies of the dataset.

        Parameters
        ----------
        delta : Optional[datetime.timedelta]
            The time delta for the tendencies.

        Returns
        -------
        Dict[str, NDArray[Any]]
            The statistical tendencies.
        """
        if delta is None:
            delta = self.frequency
        return {k: v[self.indices] for k, v in self.dataset.statistics_tendencies(delta).items()}

    def metadata_specific(self, **kwargs: Any) -> Dict[str, Any]:
        """Get the specific metadata of the dataset.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Dict[str, Any]
            The specific metadata.
        """
        return super().metadata_specific(indices=self.indices, **kwargs)

    def source(self, index: int) -> Source:
        """Get the source of the dataset.

        Parameters
        ----------
        index : int
            The index of the source.

        Returns
        -------
        Source
            The source of the dataset.
        """
        return Source(self, index, self.dataset.source(self.indices[index]))

    def tree(self) -> Node:
        """Get the tree representation of the dataset.

        Returns
        -------
        Node
            The tree representation of the dataset.
        """
        return Node(self, [self.dataset.tree()], **self.reason)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata specific to the subclass.
        """
        # return dict(indices=self.indices)
        return dict(reason=self.reason)


class Rename(Forwards):
    """Class to rename variables in a dataset."""

    def __init__(self, dataset: Dataset, rename: Dict[str, str]) -> None:
        """Initialize the Rename class.

        Parameters
        ----------
        dataset : Dataset
            The dataset to rename.
        rename : Dict[str, str]
            The mapping of old names to new names.
        """
        super().__init__(dataset)
        for n in rename:
            assert n in dataset.variables, n

        self._variables = [rename.get(v, v) for v in dataset.variables]
        self._variables_metadata = {rename.get(k, k): v for k, v in dataset.variables_metadata.items()}

        self.rename = rename

    @property
    def variables(self) -> List[str]:
        """Get the renamed variables."""
        return self._variables

    @property
    def variables_metadata(self) -> Dict[str, Any]:
        """Get the renamed variables metadata."""
        return self._variables_metadata

    @cached_property
    def name_to_index(self) -> Dict[str, int]:
        """Get the mapping of renamed variable names to indices."""
        return {k: i for i, k in enumerate(self.variables)}

    def tree(self) -> Node:
        """Get the tree representation of the dataset.

        Returns:
            Node: The tree representation of the dataset.
        """
        return Node(self, [self.forward.tree()], rename=self.rename)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the subclass.

        Returns:
            Dict[str, Any]: The metadata specific to the subclass.
        """
        return dict(rename=self.rename)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Set

import numpy as np
from numpy.typing import NDArray

from anemoi.datasets.data import MissingDateError

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import TupleIndex
from .debug import Node
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class MissingDatesFill(Forwards):
    """Class to handle filling missing dates in a dataset."""

    def __init__(self, dataset: Dataset) -> None:
        """Initialize the MissingDatesFill class.

        Parameters
        ----------
        dataset : Dataset
            The dataset with missing dates.
        """
        super().__init__(dataset)
        self._missing = set(dataset.missing)
        self._warnings: Set[int] = set()

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Get a tuple of data for the given index.

        Parameters
        ----------
        index : TupleIndex
            The index to retrieve data for.

        Returns
        -------
        NDArray[Any]
            The data for the given index.
        """
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, 0, slice(None))
        result = self._get_slice(previous)
        return apply_index_to_slices_changes(result[index], changes)

    def _get_slice(self, s: slice) -> NDArray[Any]:
        """Get a slice of data.

        Parameters
        ----------
        s : slice
            The slice to retrieve data for.

        Returns
        -------
        NDArray[Any]
            The data for the given slice.
        """
        return np.stack([self[i] for i in range(*s.indices(self._len))])

    @property
    def missing(self) -> Set[int]:
        """Get the set of missing dates."""
        return set()

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Get the data for the given index.

        Parameters
        ----------
        n : FullIndex
            The index to retrieve data for.

        Returns
        -------
        NDArray[Any]
            The data for the given index.
        """
        try:
            return self.forward[n]
        except MissingDateError:
            pass

        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        if n < 0:
            n += self._len

        a = None
        i = n
        while a is None and i >= 0:
            if i in self._missing:
                i -= 1
            else:
                a = i

        len = self._len
        b = None
        i = n
        while b is None and n < len:
            if i in self._missing:
                i += 1
            else:
                b = i

        return self._fill_missing(n, a, b)


class MissingDatesClosest(MissingDatesFill):
    """Class to handle filling missing dates using the closest available date."""

    def __init__(self, dataset: Any, closest: str) -> None:
        """Initialize the MissingDatesClosest class.

        Parameters
        ----------
        dataset : Any
            The dataset with missing dates.
        closest : str
            The strategy to use for filling missing dates ('up' or 'down').
        """
        super().__init__(dataset)
        self.closest = closest
        self._closest = {}

    def _fill_missing(self, n: int, a: Optional[int], b: Optional[int]) -> NDArray[Any]:
        """Fill the missing date at the given index.

        Parameters
        ----------
        n : int
            The index of the missing date.
        a : Optional[int]
            The previous available date index.
        b : Optional[int]
            The next available date index.

        Returns
        -------
        NDArray[Any]
            The filled data for the missing date.
        """
        if n not in self._warnings:
            LOG.warning(f"Missing date at index {n} ({self.dates[n]})")
            if abs(n - a) == abs(b - n):
                if self.closest == "up":
                    u = b
                else:
                    u = a
            else:
                if abs(n - a) < abs(b - n):
                    u = a
                else:
                    u = b
            LOG.warning(f"Using closest date {u} ({self.dates[u]})")

            self._closest[n] = u
            self._warnings.add(n)

        return self.forward[self._closest[n]]

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get metadata specific to the subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata specific to the subclass.
        """
        return {"closest": self.closest}

    def tree(self) -> Node:
        """Get the tree representation of the object.

        Returns
        -------
        Node
            The tree representation of the object.
        """
        return Node(self, [self.forward.tree()], closest=self.closest)


class MissingDatesInterpolate(MissingDatesFill):
    """Class to handle filling missing dates using interpolation."""

    def __init__(self, dataset: Any) -> None:
        """Initialize the MissingDatesInterpolate class.

        Parameters
        ----------
        dataset : Any
            The dataset with missing dates.
        """
        super().__init__(dataset)
        self._alpha = {}

    def _fill_missing(self, n: int, a: Optional[int], b: Optional[int]) -> NDArray[Any]:
        """Fill the missing date at the given index using interpolation.

        Parameters
        ----------
        n : int
            The index of the missing date.
        a : Optional[int]
            The previous available date index.
        b : Optional[int]
            The next available date index.

        Returns
        -------
        NDArray[Any]
            The filled data for the missing date.
        """
        if n not in self._warnings:
            LOG.warning(f"Missing date at index {n} ({self.dates[n]})")

            if a is None or b is None:
                raise MissingDateError(
                    f"Cannot interpolate at index {n} ({self.dates[n]}). Are the first or last date missing?"
                )

            assert a < n < b, (a, n, b)

            alpha = (n - a) / (b - a)
            assert 0 < alpha < 1, alpha

            LOG.warning(f"Interpolating between index {a} ({self.dates[a]}) and {b} ({self.dates[b]})")
            LOG.warning(f"Interpolation {1 - alpha:g} * ({self.dates[a]}) + {alpha:g} * ({self.dates[b]})")

            self._alpha[n] = alpha

            self._warnings.add(n)

        alpha = self._alpha[n]
        return self.forward[a] * (1 - alpha) + self.forward[b] * alpha

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get metadata specific to the subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata specific to the subclass.
        """
        return {}

    def tree(self) -> Node:
        """Get the tree representation of the object.

        Returns
        -------
        Node
            The tree representation of the object.
        """
        return Node(self, [self.forward.tree()])


def fill_missing_dates_factory(dataset: Any, method: str, kwargs: Dict[str, Any]) -> Dataset:
    """Factory function to create an instance of a class to fill missing dates.

    Parameters
    ----------
    dataset : Any
        The dataset with missing dates.
    method : str
        The method to use for filling missing dates ('closest' or 'interpolate').
    kwargs : Dict[str, Any]
        Additional arguments for the method.

    Returns
    -------
    Dataset
        An instance of a class to fill missing dates.
    """
    if method == "closest":
        closest = kwargs.get("closest", "up")
        return MissingDatesClosest(dataset, closest=closest)

    if method == "interpolate":
        return MissingDatesInterpolate(dataset)

    raise ValueError(f"Invalid `fill_missing_dates` method '{method}'")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import fnmatch
import os
from typing import Any

import tqdm

from . import Command


class GribIndexCmd(Command):
    internal = True
    timestamp = True

    def add_arguments(self, command_parser: Any) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : Any
            The command parser to which arguments are added.
        """

        from anemoi.datasets.create.sources.grib_index import KEYS

        command_parser.add_argument(
            "--index",
            help="Path to the index file to create or update",
            required=True,
        )

        command_parser.add_argument(
            "--overwrite",
            action="store_true",
            help="Overwrite the index file if it exists (default is to update)",
        )

        command_parser.add_argument(
            "--match",
            help="Give a glob pattern to match files (default: *.grib)",
            default="*.grib",
        )

        command_parser.add_argument(
            "--keys",
            help="GRIB keys to add to the index, separated by commas. If the list starts with a +, the keys are added to default list.",
            default=",".join(KEYS),
        )

        command_parser.add_argument(
            "--flavour",
            help="GRIB flavour file (yaml or json)",
        )

        command_parser.add_argument("paths", nargs="+", help="Paths to scan")

    def run(self, args: Any) -> None:
        """Execute the scan command.

        Parameters
        ----------
        args : Any
            The arguments passed to the command.
        """

        def match(path: str) -> bool:
            """Check if a path matches the given pattern.

            Parameters
            ----------
            path : str
                The path to check.

            Returns
            -------
            bool
                True if the path matches, False otherwise.
            """
            return fnmatch.fnmatch(path, args.match)

        from anemoi.datasets.create.sources.grib_index import GribIndex

        index = GribIndex(
            args.index,
            keys=args.keys,
            update=True,
            overwrite=args.overwrite,
            flavour=args.flavour,
        )

        paths = []
        for path in args.paths:
            if os.path.isfile(path):
                paths.append(path)
            else:
                for root, _, files in os.walk(path):
                    for file in files:
                        full = os.path.join(root, file)
                        paths.append(full)

        for path in tqdm.tqdm(paths, leave=False):
            if match(path):
                index.add_grib_file(path)


command = GribIndexCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import Hashable
from typing import Optional
from typing import Tuple

import xarray as xr
from anemoi.utils.config import DotDict

from .coordinates import Coordinate
from .coordinates import DateCoordinate
from .coordinates import EnsembleCoordinate
from .coordinates import LatitudeCoordinate
from .coordinates import LevelCoordinate
from .coordinates import LongitudeCoordinate
from .coordinates import ScalarCoordinate
from .coordinates import StepCoordinate
from .coordinates import TimeCoordinate
from .coordinates import UnsupportedCoordinate
from .coordinates import XCoordinate
from .coordinates import YCoordinate
from .coordinates import is_scalar
from .grid import Grid
from .grid import MeshedGrid
from .grid import MeshProjectionGrid
from .grid import UnstructuredGrid
from .grid import UnstructuredProjectionGrid

LOG = logging.getLogger(__name__)

# CoordinateAttributes = namedtuple("CoordinateAttributes", ["axis", "name", "long_name", "standard_name", "units"])


class CoordinateAttributes(DotDict):
    pass


class CoordinateGuesser(ABC):
    """Class to guess the type of coordinates in a dataset."""

    def __init__(self, ds: xr.Dataset) -> None:
        """Initializes the CoordinateGuesser.

        Parameters
        ----------
        ds : xr.Dataset
            The dataset to guess coordinates from.
        """
        self.ds = ds
        self._coordinate_cache: Dict[Hashable, Coordinate] = {}
        self._grid_cache: Dict[Hashable, Grid] = {}

    @classmethod
    def from_flavour(cls, ds: xr.Dataset, flavour: Optional[Dict[str, Any]]) -> "CoordinateGuesser":
        """Creates a CoordinateGuesser from a flavour.

        Parameters
        ----------
        ds : xr.Dataset
            The dataset to guess coordinates from.
        flavour : Optional[Dict[str, Any]]
            The flavour to use for guessing.

        Returns
        -------
        CoordinateGuesser
            The created CoordinateGuesser.
        """
        if flavour is None:
            return DefaultCoordinateGuesser(ds)
        else:
            return FlavourCoordinateGuesser(ds, flavour)

    def guess(self, c: xr.DataArray, coord: Hashable) -> Coordinate:
        """Guesses the type of a coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to guess.
        coord : Hashable
            The name of the coordinate.

        Returns
        -------
        Coordinate
            The guessed coordinate type.
        """
        if coord not in self._coordinate_cache:
            self._coordinate_cache[coord] = self._guess(c, coord)
        return self._coordinate_cache[coord]

    def _guess(self, coordinate: xr.DataArray, coord: Hashable) -> Coordinate:
        """Internal method to guess the type of a coordinate.

        Parameters
        ----------
        coordinate : xr.DataArray
            The coordinate to guess.
        coord : Hashable
            The name of the coordinate.

        Returns
        -------
        Coordinate
            The guessed coordinate type.
        """
        name = coordinate.name
        standard_name = getattr(coordinate, "standard_name", "").lower()
        axis = getattr(coordinate, "axis", "")
        long_name = getattr(coordinate, "long_name", "").lower()
        units = getattr(coordinate, "units", "")

        attributes = CoordinateAttributes(
            axis=axis,
            name=name,
            long_name=long_name,
            standard_name=standard_name,
            units=units,
        )

        d: Optional[Coordinate] = None

        d = self._is_longitude(coordinate, attributes)
        if d is not None:
            return d

        d = self._is_latitude(coordinate, attributes)
        if d is not None:
            return d

        d = self._is_x(coordinate, attributes)
        if d is not None:
            return d

        d = self._is_y(coordinate, attributes)
        if d is not None:
            return d

        d = self._is_time(coordinate, attributes)
        if d is not None:
            return d

        d = self._is_step(coordinate, attributes)
        if d is not None:
            return d

        d = self._is_date(coordinate, attributes)
        if d is not None:
            return d

        d = self._is_level(coordinate, attributes)
        if d is not None:
            return d

        d = self._is_number(coordinate, attributes)
        if d is not None:
            return d

        if coordinate.shape in ((1,), tuple()):
            return ScalarCoordinate(coordinate)

        LOG.warning(
            f"Coordinate {coord} not supported\n{axis=}, {name=},"
            f" {long_name=}, {standard_name=}, units\n\n{coordinate}\n\n{type(coordinate.values)} {coordinate.shape}"
        )

        return UnsupportedCoordinate(coordinate)

    def grid(self, coordinates: Any, variable: Any) -> Any:
        """Determines the grid type for the given coordinates and variable.

        Parameters
        ----------
        coordinates : Any
            The coordinates to determine the grid from.
        variable : Any
            The variable to determine the grid from.

        Returns
        -------
        Any
            The determined grid type.
        """
        lat = [c for c in coordinates if c.is_lat]
        lon = [c for c in coordinates if c.is_lon]

        if len(lat) == 1 and len(lon) == 1:
            return self._lat_lon_provided(lat, lon, variable)

        x = [c for c in coordinates if c.is_x]
        y = [c for c in coordinates if c.is_y]

        if len(x) == 1 and len(y) == 1:
            return self._x_y_provided(x, y, variable)

        raise NotImplementedError(f"Cannot establish grid {coordinates}")

    def _check_dims(self, variable: Any, x_or_lon: Any, y_or_lat: Any) -> Tuple[Any, bool]:
        """Checks the dimensions of the variable against the coordinates.

        Parameters
        ----------
        variable : Any
            The variable to check.
        x_or_lon : Any
            The x or longitude coordinate.
        y_or_lat : Any
            The y or latitude coordinate.

        Returns
        -------
        Tuple[Any, bool]
            The checked dimensions and a flag indicating if the grid is unstructured.
        """
        x_dims = set(x_or_lon.variable.dims)
        y_dims = set(y_or_lat.variable.dims)
        variable_dims = set(variable.dims)

        if not (x_dims <= variable_dims) or not (y_dims <= variable_dims):
            raise ValueError(
                f"Dimensions do not match {variable.name}{variable.dims} !="
                f" {x_or_lon.name}{x_or_lon.variable.dims} and {y_or_lat.name}{y_or_lat.variable.dims}"
            )

        variable_dims = tuple(v for v in variable.dims if v in (x_dims | y_dims))
        if x_dims == y_dims:
            # It's unstructured
            return variable_dims, True

        if len(x_dims) == 1 and len(y_dims) == 1:
            # It's a mesh
            return variable_dims, False

        raise ValueError(
            f"Cannot establish grid for {variable.name}{variable.dims},"
            f" {x_or_lon.name}{x_or_lon.variable.dims},"
            f" {y_or_lat.name}{y_or_lat.variable.dims}"
        )

    def _lat_lon_provided(self, lat: Any, lon: Any, variable: Any) -> Any:
        """Determines the grid type when latitude and longitude are provided.

        Parameters
        ----------
        lat : Any
            The latitude coordinate.
        lon : Any
            The longitude coordinate.
        variable : Any
            The variable to determine the grid from.

        Returns
        -------
        Any
            The determined grid type.
        """
        lat = lat[0]
        lon = lon[0]

        dim_vars, unstructured = self._check_dims(variable, lon, lat)

        if (lat.name, lon.name, dim_vars) in self._grid_cache:
            return self._grid_cache[(lat.name, lon.name, dim_vars)]

        grid: Grid = UnstructuredGrid(lat, lon, dim_vars) if unstructured else MeshedGrid(lat, lon, dim_vars)

        self._grid_cache[(lat.name, lon.name, dim_vars)] = grid

        return grid

    def _x_y_provided(self, x: Any, y: Any, variable: Any) -> Any:
        """Determines the grid type when x and y coordinates are provided.

        Parameters
        ----------
        x : Any
            The x coordinate.
        y : Any
            The y coordinate.
        variable : Any
            The variable to determine the grid from.

        Returns
        -------
        Any
            The determined grid type.
        """
        x = x[0]
        y = y[0]

        dim_vars, unstructured = self._check_dims(variable, x, y)

        if (x.name, y.name, dim_vars) in self._grid_cache:
            return self._grid_cache[(x.name, y.name, dim_vars)]

        grid_mapping = variable.attrs.get("grid_mapping", None)
        if grid_mapping is not None:
            print(f"grid_mapping: {grid_mapping}")
            print(self.ds[grid_mapping])

        if grid_mapping is None:
            LOG.warning(f"No 'grid_mapping' attribute provided for '{variable.name}'")
            LOG.warning("Trying to guess...")

            PROBE = {
                "prime_meridian_name",
                "reference_ellipsoid_name",
                "crs_wkt",
                "horizontal_datum_name",
                "semi_major_axis",
                "spatial_ref",
                "inverse_flattening",
                "semi_minor_axis",
                "geographic_crs_name",
                "GeoTransform",
                "grid_mapping_name",
                "longitude_of_prime_meridian",
            }
            candidate = None
            for v in self.ds.variables:
                var = self.ds[v]
                if not is_scalar(var):
                    continue

                if PROBE.intersection(var.attrs.keys()):
                    if candidate:
                        raise ValueError(f"Multiple candidates for 'grid_mapping': {candidate} and {v}")
                    candidate = v

            if candidate:
                LOG.warning(f"Using '{candidate}' as 'grid_mapping'")
                grid_mapping = candidate
            else:
                LOG.warning("Could not fine a candidate for 'grid_mapping'")

        if grid_mapping is None:
            if "crs" in self.ds[variable].attrs:
                grid_mapping = self.ds[variable].attrs["crs"]
                LOG.warning(f"Using CRS {grid_mapping} from variable '{variable.name}' attributes")

        if grid_mapping is None:
            if "crs" in self.ds.attrs:
                grid_mapping = self.ds.attrs["crs"]
                LOG.warning(f"Using CRS {grid_mapping} from global attributes")

        grid: Optional[Grid] = None
        if grid_mapping is not None:

            grid_mapping = dict(self.ds[grid_mapping].attrs)

            if unstructured:
                grid = UnstructuredProjectionGrid(x, y, grid_mapping)
            else:
                grid = MeshProjectionGrid(x, y, grid_mapping)

        if grid is not None:
            self._grid_cache[(x.name, y.name, dim_vars)] = grid
            return grid

        LOG.error("Could not fine a candidate for 'grid_mapping'")
        raise NotImplementedError(f"Unstructured grid {x.name} {y.name}")

    @abstractmethod
    def _is_longitude(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LongitudeCoordinate]:
        """Checks if the coordinate is a longitude.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LongitudeCoordinate]
            The LongitudeCoordinate if matched, else None.
        """
        pass

    @abstractmethod
    def _is_latitude(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LatitudeCoordinate]:
        """Checks if the coordinate is a latitude.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LatitudeCoordinate]
            The LatitudeCoordinate if matched, else None.
        """
        pass

    @abstractmethod
    def _is_x(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[XCoordinate]:
        """Checks if the coordinate is an x coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[XCoordinate]
            The XCoordinate if matched, else None.
        """
        pass

    @abstractmethod
    def _is_y(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[YCoordinate]:
        """Checks if the coordinate is a y coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[YCoordinate]
            The YCoordinate if matched, else None.
        """
        pass

    @abstractmethod
    def _is_time(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[TimeCoordinate]:
        """Checks if the coordinate is a time coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[TimeCoordinate]
            The TimeCoordinate if matched, else None.
        """
        pass

    @abstractmethod
    def _is_date(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[DateCoordinate]:
        """Checks if the coordinate is a date coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[DateCoordinate]
            The DateCoordinate if matched, else None.
        """
        pass

    @abstractmethod
    def _is_step(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[StepCoordinate]:
        """Checks if the coordinate is a step coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[StepCoordinate]
            The StepCoordinate if matched, else None.
        """
        pass

    @abstractmethod
    def _is_level(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LevelCoordinate]:
        """Checks if the coordinate is a level coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LevelCoordinate]
            The LevelCoordinate if matched, else None.
        """
        pass

    @abstractmethod
    def _is_number(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[EnsembleCoordinate]:
        """Checks if the coordinate is an ensemble coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[EnsembleCoordinate]
            The EnsembleCoordinate if matched, else None.
        """
        pass


class DefaultCoordinateGuesser(CoordinateGuesser):
    """Default implementation of CoordinateGuesser."""

    def __init__(self, ds: xr.Dataset) -> None:
        """Initializes the DefaultCoordinateGuesser.

        Parameters
        ----------
        ds : xr.Dataset
            The dataset to guess coordinates from.
        """
        super().__init__(ds)

    def _is_longitude(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LongitudeCoordinate]:
        """Checks if the coordinate is a longitude.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LongitudeCoordinate]
            The LongitudeCoordinate if matched, else None.
        """

        # https://cfconventions.org/Data/cf-conventions/cf-conventions-1.12/cf-conventions.html#longitude-coordinate

        if attributes.standard_name == "longitude":
            return LongitudeCoordinate(c)

        if attributes.long_name == "longitude" and attributes.units == "degrees_east":
            return LongitudeCoordinate(c)

        if attributes.name == "longitude":  # WeatherBench
            return LongitudeCoordinate(c)

        return None

    def _is_latitude(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LatitudeCoordinate]:
        """Checks if the coordinate is a latitude.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LatitudeCoordinate]
            The LatitudeCoordinate if matched, else None.
        """

        # https://cfconventions.org/Data/cf-conventions/cf-conventions-1.12/cf-conventions.html#latitude-coordinate

        if attributes.standard_name == "latitude":
            return LatitudeCoordinate(c)

        if attributes.long_name == "latitude" and attributes.units == "degrees_north":
            return LatitudeCoordinate(c)

        if attributes.name == "latitude":  # WeatherBench
            return LatitudeCoordinate(c)

        return None

    def _is_x(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[XCoordinate]:
        """Checks if the coordinate is an x coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[XCoordinate]
            The XCoordinate if matched, else None.
        """
        if attributes.standard_name in ["projection_x_coordinate", "grid_longitude"]:
            return XCoordinate(c)

        if attributes.name == "x":
            return XCoordinate(c)

        return None

    def _is_y(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[YCoordinate]:
        """Checks if the coordinate is a y coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[YCoordinate]
            The YCoordinate if matched, else None.
        """
        if attributes.standard_name in ["projection_y_coordinate", "grid_latitude"]:
            return YCoordinate(c)

        if attributes.name == "y":
            return YCoordinate(c)

        return None

    def _is_time(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[TimeCoordinate]:
        """Checks if the coordinate is a time coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[TimeCoordinate]
            The TimeCoordinate if matched, else None.
        """
        if attributes.standard_name == "time":
            return TimeCoordinate(c)

        if attributes.name == "time" and attributes.standard_name != "forecast_reference_time":
            return TimeCoordinate(c)

        return None

    def _is_date(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[DateCoordinate]:
        """Checks if the coordinate is a date coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[DateCoordinate]
            The DateCoordinate if matched, else None.
        """
        if attributes.standard_name == "forecast_reference_time":
            return DateCoordinate(c)

        if attributes.name == "forecast_reference_time":
            return DateCoordinate(c)

        return None

    def _is_step(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[StepCoordinate]:
        """Checks if the coordinate is a step coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[StepCoordinate]
            The StepCoordinate if matched, else None.
        """
        if attributes.standard_name == "forecast_period":
            return StepCoordinate(c)

        if attributes.long_name == "time elapsed since the start of the forecast":
            return StepCoordinate(c)

        if attributes.name == "prediction_timedelta":  # WeatherBench
            return StepCoordinate(c)

        return None

    def _is_level(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LevelCoordinate]:
        """Checks if the coordinate is a level coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LevelCoordinate]
            The LevelCoordinate if matched, else None.
        """
        if attributes.standard_name == "atmosphere_hybrid_sigma_pressure_coordinate":
            return LevelCoordinate(c, "ml")

        if attributes.long_name == "height" and attributes.units == "m":
            return LevelCoordinate(c, "height")

        if attributes.standard_name == "air_pressure" and attributes.units == "hPa":
            return LevelCoordinate(c, "pl")

        if attributes.name == "level":
            return LevelCoordinate(c, "pl")

        if attributes.name == "vertical" and attributes.units == "hPa":
            return LevelCoordinate(c, "pl")

        if attributes.standard_name == "depth":
            return LevelCoordinate(c, "depth")

        if attributes.name == "vertical" and attributes.units == "hPa":
            return LevelCoordinate(c, "pl")

        return None

    def _is_number(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[EnsembleCoordinate]:
        """Checks if the coordinate is an ensemble coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[EnsembleCoordinate]
            The EnsembleCoordinate if matched, else None.
        """
        if attributes.name in ("realization", "number"):
            return EnsembleCoordinate(c)

        return None


class FlavourCoordinateGuesser(CoordinateGuesser):
    """Implementation of CoordinateGuesser that uses a flavour for guessing."""

    def __init__(self, ds: xr.Dataset, flavour: Dict[str, Any]) -> None:
        """Initializes the FlavourCoordinateGuesser.

        Parameters
        ----------
        ds : xr.Dataset
            The dataset to guess coordinates from.
        flavour : Dict[str, Any]
            The flavour to use for guessing.
        """
        super().__init__(ds)
        self.flavour = flavour

    def _match(self, c: xr.DataArray, key: str, attributes: CoordinateAttributes) -> Optional[Dict[str, Any]]:
        """Matches the coordinate against the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to match.
        key : str
            The key to match in the flavour rules.
        attributes : CoordinateAttributes
            The values to match against.

        Returns
        -------
        Optional[Dict[str, Any]]
            The matched rule if any, else None.
        """
        if key not in self.flavour["rules"]:
            return None

        rules = self.flavour["rules"][key]

        if not isinstance(rules, list):
            rules = [rules]

        for rule in rules:
            ok = True
            for k, v in rule.items():
                if isinstance(v, str) and attributes.get(k) != v:
                    ok = False
            if ok:
                return rule

        return None

    def _is_longitude(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LongitudeCoordinate]:
        """Checks if the coordinate is a longitude using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LongitudeCoordinate]
            The LongitudeCoordinate if matched, else None.
        """
        if self._match(c, "longitude", attributes):
            return LongitudeCoordinate(c)

        return None

    def _is_latitude(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LatitudeCoordinate]:
        """Checks if the coordinate is a latitude using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LatitudeCoordinate]
            The LatitudeCoordinate if matched, else None.
        """
        if self._match(c, "latitude", attributes):
            return LatitudeCoordinate(c)

        return None

    def _is_x(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[XCoordinate]:
        """Checks if the coordinate is an x coordinate using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[XCoordinate]
            The XCoordinate if matched, else None.
        """
        if self._match(c, "x", attributes):
            return XCoordinate(c)

        return None

    def _is_y(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[YCoordinate]:
        """Checks if the coordinate is a y coordinate using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[YCoordinate]
            The YCoordinate if matched, else None.
        """
        if self._match(c, "y", attributes):
            return YCoordinate(c)

        return None

    def _is_time(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[TimeCoordinate]:
        """Checks if the coordinate is a time coordinate using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[TimeCoordinate]
            The TimeCoordinate if matched, else None.
        """
        if self._match(c, "time", attributes):
            return TimeCoordinate(c)

        return None

    def _is_step(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[StepCoordinate]:
        """Checks if the coordinate is a step coordinate using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[StepCoordinate]
            The StepCoordinate if matched, else None.
        """
        if self._match(c, "step", attributes):
            return StepCoordinate(c)

        return None

    def _is_date(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[DateCoordinate]:
        """Checks if the coordinate is a date coordinate using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[DateCoordinate]
            The DateCoordinate if matched, else None.
        """
        if self._match(c, "date", attributes):
            return DateCoordinate(c)

        return None

    def _is_level(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[LevelCoordinate]:
        """Checks if the coordinate is a level coordinate using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[LevelCoordinate]
            The LevelCoordinate if matched, else None.
        """
        rule = self._match(c, "level", attributes)
        if rule:
            # assert False, rule
            return LevelCoordinate(
                c,
                self._levtype(c, attributes),
            )

        return None

    def _levtype(self, c: xr.DataArray, attributes: CoordinateAttributes) -> str:
        """Determines the level type for the coordinate.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        str
            The level type.
        """
        if "levtype" in self.flavour:
            return self.flavour["levtype"]

        raise NotImplementedError(f"levtype for {c=}")

    def _is_number(self, c: xr.DataArray, attributes: CoordinateAttributes) -> Optional[EnsembleCoordinate]:
        """Checks if the coordinate is an ensemble coordinate using the flavour rules.

        Parameters
        ----------
        c : xr.DataArray
            The coordinate to check.
        attributes : CoordinateAttributes
            The attributes of the coordinate.

        Returns
        -------
        Optional[EnsembleCoordinate]
            The EnsembleCoordinate if matched, else None.
        """
        if self._match(c, "number", attributes):
            return EnsembleCoordinate(c)

        return None



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from abc import ABC
from abc import abstractmethod

import earthkit.data as ekd

from anemoi.datasets.create.typing import DateList


class Source(ABC):
    """Represents a data source with a given context."""

    emoji = "📦"  # For tracing

    def __init__(self, context: any, *args: tuple, **kwargs: dict):
        """Initialise the source.
        Parameters
        ----------
        context : Any
            The context for the data source.
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Additional keyword arguments.
        """
        self.context = context

    @abstractmethod
    def execute(self, dates: DateList) -> ekd.FieldList:
        """Execute the filter.

        Parameters
        ----------
        dates : DateList
            The input dates.

        Returns
        -------
        ekd.FieldList
            The output data.
        """

        pass



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import glob
import hashlib
import json
import logging
import os
import pickle
import shutil
import socket
from typing import Any
from typing import Iterator
from typing import Tuple

import numpy as np
from anemoi.utils.provenance import gather_provenance_info

LOG = logging.getLogger(__name__)


class PersistentDict:
    """A dictionary-like object that persists its contents to disk using pickle files.

    Attributes
    ----------
    version : int
        The version of the PersistentDict.
    dirname : str
        The directory where the data is stored.
    name : str
        The name of the directory.
    ext : str
        The extension of the directory.
    """

    version = 3

    # Used in parrallel, during data loading,
    # to write data in pickle files.
    def __init__(self, directory: str, create: bool = True):
        """Initialize the PersistentDict.

        Parameters
        ----------
        directory : str
            The directory where the data will be stored.
        create : bool, optional
            Whether to create the directory if it doesn't exist.
        """
        self.dirname = directory
        self.name, self.ext = os.path.splitext(os.path.basename(self.dirname))
        if create:
            self.create()

    def create(self) -> None:
        """Create the directory if it doesn't exist."""
        os.makedirs(self.dirname, exist_ok=True)

    def delete(self) -> None:
        """Delete the directory and its contents."""
        try:
            shutil.rmtree(self.dirname)
        except FileNotFoundError:
            pass

    def __str__(self) -> str:
        """Return a string representation of the PersistentDict."""
        return f"{self.__class__.__name__}({self.dirname})"

    def items(self) -> Iterator[Any]:
        """Yield items stored in the directory.

        Yields
        ------
        Iterator[Any]
            An iterator over the items.
        """
        # use glob to read all pickles
        files = glob.glob(self.dirname + "/*.pickle")
        LOG.debug(f"Reading {self.name} data, found {len(files)} files in {self.dirname}")
        assert len(files) > 0, f"No files found in {self.dirname}"
        for f in files:
            with open(f, "rb") as f:
                yield pickle.load(f)

    def add_provenance(self, **kwargs: Any) -> None:
        """Add provenance information to the directory.

        Parameters
        ----------
        **kwargs : Any
            Additional provenance information.
        """
        path = os.path.join(self.dirname, "provenance.json")
        if os.path.exists(path):
            return
        out = dict(provenance=gather_provenance_info(), **kwargs)
        with open(path, "w") as f:
            json.dump(out, f)

    def add(self, elt: Any, *, key: Any) -> None:
        """Add an element to the PersistentDict.

        Parameters
        ----------
        elt : Any
            The element to add.
        key : Any
            The key associated with the element.
        """
        self[key] = elt

    def __setitem__(self, key: Any, elt: Any) -> None:
        """Set an item in the PersistentDict.

        Parameters
        ----------
        key : Any
            The key associated with the element.
        elt : Any
            The element to set.
        """
        h = hashlib.sha256(str(key).encode("utf-8")).hexdigest()
        path = os.path.join(self.dirname, f"{h}.pickle")

        if os.path.exists(path):
            LOG.warning(f"{path} already exists")

        tmp_path = path + f".tmp-{os.getpid()}-on-{socket.gethostname()}"
        with open(tmp_path, "wb") as f:
            pickle.dump((key, elt), f)
        shutil.move(tmp_path, path)

        LOG.debug(f"Written {self.name} data for len {key} in {path}")

    def flush(self) -> None:
        """Flush the PersistentDict (no-op)."""
        pass


class BufferedPersistentDict(PersistentDict):
    """A buffered version of PersistentDict that stores elements in memory before persisting them to disk.

    Attributes
    ----------
    buffer_size : int
        The size of the buffer.
    elements : list
        The list of elements in the buffer.
    keys : list
        The list of keys in the buffer.
    storage : PersistentDict
        The underlying PersistentDict used for storage.
    """

    def __init__(self, buffer_size: int = 1000, **kwargs: Any):
        """Initialize the BufferedPersistentDict.

        Parameters
        ----------
        buffer_size : int, optional
            The size of the buffer.
        **kwargs : Any
            Additional arguments for PersistentDict.
        """
        self.buffer_size = buffer_size
        self.elements = []
        self.keys = []
        self.storage = PersistentDict(**kwargs)

    def add(self, elt: Any, *, key: Any) -> None:
        """Add an element to the BufferedPersistentDict.

        Parameters
        ----------
        elt : Any
            The element to add.
        key : Any
            The key associated with the element.
        """
        self.elements.append(elt)
        self.keys.append(key)
        if len(self.keys) > self.buffer_size:
            self.flush()

    def flush(self) -> None:
        """Flush the buffer and store the elements in PersistentDict."""
        k = sorted(self.keys)
        self.storage.add(self.elements, key=k)
        self.elements = []
        self.keys = []

    def items(self) -> Iterator[Tuple[Any, Any]]:
        """Yield items stored in the BufferedPersistentDict.

        Yields
        ------
        Iterator[Tuple[Any, Any]]
            An iterator over the items.
        """
        for keys, elements in self.storage.items():
            for key, elt in zip(keys, elements):
                yield key, elt

    def delete(self) -> None:
        """Delete the storage directory and its contents."""
        self.storage.delete()

    def create(self) -> None:
        """Create the storage directory if it doesn't exist."""
        self.storage.create()


def build_storage(directory: str, create: bool = True) -> BufferedPersistentDict:
    """Build a BufferedPersistentDict storage.

    Parameters
    ----------
    directory : str
        The directory where the data will be stored.
    create : bool, optional
        Whether to create the directory if it doesn't exist.

    Returns
    -------
    BufferedPersistentDict
        The created BufferedPersistentDict.
    """
    return BufferedPersistentDict(directory=directory, create=create)


if __name__ == "__main__":
    N = 3
    P = 2
    directory = "h"
    p = PersistentDict(directory=directory)
    print(p)
    assert os.path.exists(directory)
    import numpy as np

    arrs = [np.random.randint(1, 101, size=(P,)) for _ in range(N)]
    dates = [np.array([np.datetime64(f"2021-01-0{_+1}") + np.timedelta64(i, "h") for i in range(P)]) for _ in range(N)]

    print()
    print("Writing the data")
    for i in range(N):
        _arr = arrs[i]
        _dates = dates[i]
        print(f"Writing : {i=}, {_arr=} {_dates=}")
        p[_dates] = (i, _arr)

    print()
    print("Reading the data back")

    p = PersistentDict(directory="h")
    for _dates, (i, _arr) in p.items():
        print(f"{i=}, {_arr=}, {_dates=}")

        assert np.allclose(_arr, arrs[i])

        assert len(_dates) == len(dates[i])
        for a, b in zip(_dates, dates[i]):
            assert a == b



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
import shutil
from typing import Any
from typing import Optional

import numpy as np
import zarr
from numpy.typing import NDArray

LOG = logging.getLogger(__name__)


def add_zarr_dataset(
    *,
    name: str,
    dtype: np.dtype = None,
    fill_value: np.generic = None,
    zarr_root: zarr.Group,
    shape: tuple[int, ...] = None,
    array: NDArray[Any] = None,
    overwrite: bool = True,
    dimensions: tuple[str, ...] = None,
    **kwargs,
) -> zarr.Array:
    """Add a dataset to a Zarr group.

    Parameters
    ----------
    name : str
        Name of the dataset.
    dtype : np.dtype, optional
        Data type of the dataset.
    fill_value : np.generic, optional
        Fill value for the dataset.
    zarr_root : zarr.Group
        Root Zarr group.
    shape : tuple[int, ...], optional
        Shape of the dataset.
    array : NDArray[Any], optional
        Array to initialize the dataset with.
    overwrite : bool
        Whether to overwrite existing dataset.
    dimensions : tuple[str, ...]
        Dimensions of the dataset.
    **kwargs
        Additional arguments for Zarr dataset creation.

    Returns
    -------
    zarr.Array
        The created Zarr array.
    """
    assert dimensions is not None, "Please pass dimensions to add_zarr_dataset."
    assert isinstance(dimensions, (tuple, list))

    if dtype is None:
        assert array is not None, (name, shape, array, dtype, zarr_root)
        dtype = array.dtype

    if shape is None:
        assert array is not None, (name, shape, array, dtype, zarr_root)
        shape = array.shape

    if array is not None:
        assert array.shape == shape, (array.shape, shape)
        a = zarr_root.create_dataset(
            name,
            shape=shape,
            dtype=dtype,
            overwrite=overwrite,
            **kwargs,
        )
        a[...] = array
        a.attrs["_ARRAY_DIMENSIONS"] = dimensions
        return a

    if "fill_value" not in kwargs:
        if str(dtype).startswith("float") or str(dtype).startswith("numpy.float"):
            kwargs["fill_value"] = np.nan
        elif str(dtype).startswith("datetime64") or str(dtype).startswith("numpy.datetime64"):
            kwargs["fill_value"] = np.datetime64("NaT")
        # elif str(dtype).startswith("timedelta64") or str(dtype).startswith(
        #    "numpy.timedelta64"
        # ):
        #    kwargs["fill_value"] = np.timedelta64("NaT")
        elif str(dtype).startswith("int") or str(dtype).startswith("numpy.int"):
            kwargs["fill_value"] = 0
        elif str(dtype).startswith("bool") or str(dtype).startswith("numpy.bool"):
            kwargs["fill_value"] = False
        else:
            raise ValueError(f"No fill_value for dtype={dtype}")

    a = zarr_root.create_dataset(
        name,
        shape=shape,
        dtype=dtype,
        overwrite=overwrite,
        **kwargs,
    )
    a.attrs["_ARRAY_DIMENSIONS"] = dimensions
    return a


class ZarrBuiltRegistry:
    """A class to manage the creation and access of Zarr datasets."""

    name_lengths = "lengths"
    name_flags = "flags"
    lengths = None
    flags = None
    z = None

    def __init__(self, path: str, synchronizer_path: Optional[str] = None, use_threads: bool = False):
        """Initialize the ZarrBuiltRegistry.

        Parameters
        ----------
        path : str
            Path to the Zarr store.
        synchronizer_path : Optional[str], optional
            Path to the synchronizer.
        use_threads : bool
            Whether to use thread-based synchronization.
        """
        import zarr

        assert isinstance(path, str), path
        self.zarr_path = path

        if use_threads:
            self.synchronizer = zarr.ThreadSynchronizer()
            self.synchronizer_path = None
        else:
            if synchronizer_path is None:
                synchronizer_path = self.zarr_path + ".sync"
            self.synchronizer_path = synchronizer_path
            self.synchronizer = zarr.ProcessSynchronizer(self.synchronizer_path)

    def clean(self) -> None:
        """Clean up the synchronizer path."""
        if self.synchronizer_path is not None:
            try:
                shutil.rmtree(self.synchronizer_path)
            except FileNotFoundError:
                pass

    def _open_write(self) -> zarr.Group:
        """Open the Zarr store in write mode."""
        import zarr

        return zarr.open(self.zarr_path, mode="r+", synchronizer=self.synchronizer)

    def _open_read(self, sync: bool = True) -> zarr.Group:
        """Open the Zarr store in read mode.

        Parameters
        ----------
        sync : bool
            Whether to use synchronization.

        Returns
        -------
        zarr.Group
            The opened Zarr group.
        """
        import zarr

        if sync:
            return zarr.open(self.zarr_path, mode="r", synchronizer=self.synchronizer)
        else:
            return zarr.open(self.zarr_path, mode="r")

    def new_dataset(self, *args, **kwargs) -> None:
        """Create a new dataset in the Zarr store.

        Parameters
        ----------
        *args
            Positional arguments for dataset creation.
        **kwargs
            Keyword arguments for dataset creation.
        """
        z = self._open_write()
        zarr_root = z["_build"]
        add_zarr_dataset(*args, zarr_root=zarr_root, overwrite=True, dimensions=("tmp",), **kwargs)

    def add_to_history(self, action: str, **kwargs) -> None:
        """Add an action to the history attribute of the Zarr store.

        Parameters
        ----------
        action : str
            The action to record.
        **kwargs
            Additional information about the action.
        """
        new = dict(
            action=action,
            timestamp=datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None).isoformat(),
        )
        new.update(kwargs)

        z = self._open_write()
        history = z.attrs.get("history", [])
        history.append(new)
        z.attrs["history"] = history

    def get_lengths(self) -> list[int]:
        """Get the lengths dataset.

        Returns
        -------
        list[int]
            The lengths dataset.
        """
        z = self._open_read()
        return list(z["_build"][self.name_lengths][:])

    def get_flags(self, **kwargs) -> list[bool]:
        """Get the flags dataset.

        Parameters
        ----------
        **kwargs
            Additional arguments for reading the dataset.

        Returns
        -------
        list[bool]
            The flags dataset.
        """
        z = self._open_read(**kwargs)
        return list(z["_build"][self.name_flags][:])

    def get_flag(self, i: int) -> bool:
        """Get a specific flag.

        Parameters
        ----------
        i : int
            Index of the flag.

        Returns
        -------
        bool
            The flag value.
        """
        z = self._open_read()
        return z["_build"][self.name_flags][i]

    def set_flag(self, i: int, value: bool = True) -> None:
        """Set a specific flag.

        Parameters
        ----------
        i : int
            Index of the flag.
        value : bool
            Value to set the flag to.
        """
        z = self._open_write()
        z.attrs["latest_write_timestamp"] = (
            datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None).isoformat()
        )
        z["_build"][self.name_flags][i] = value

    def ready(self) -> bool:
        """Check if all flags are set.

        Returns
        -------
        bool
            True if all flags are set, False otherwise.
        """
        return all(self.get_flags())

    def create(self, lengths: list[int], overwrite: bool = False) -> None:
        """Create the lengths and flags datasets.

        Parameters
        ----------
        lengths : list[int]
            Lengths to initialize the dataset with.
        overwrite : bool
            Whether to overwrite existing datasets.
        """
        self.new_dataset(name=self.name_lengths, array=np.array(lengths, dtype="i4"))
        self.new_dataset(name=self.name_flags, array=np.array([False] * len(lengths), dtype=bool))
        self.add_to_history("initialised")

    def reset(self, lengths: list[int]) -> None:
        """Reset the lengths and flags datasets.

        Parameters
        ----------
        lengths : list[int]
            Lengths to initialize the dataset with.
        """
        return self.create(lengths, overwrite=True)

    def add_provenance(self, name: str) -> None:
        """Add provenance information to the Zarr store.

        Parameters
        ----------
        name : str
            Name of the provenance attribute.
        """
        z = self._open_write()

        if name in z.attrs:
            return

        from anemoi.utils.provenance import gather_provenance_info

        z.attrs[name] = gather_provenance_info()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from anemoi.datasets.compute.recentre import recentre as _recentre

from .legacy import legacy_source
from .mars import mars


def to_list(x: Union[list, tuple, str]) -> List:
    """Converts the input to a list. If the input is a string, it splits it by '/'.

    Parameters
    ----------
    x : Union[list, tuple, str]
        The input to convert.

    Returns
    -------
    list
        The converted list.
    """
    if isinstance(x, (list, tuple)):
        return x
    if isinstance(x, str):
        return x.split("/")
    return [x]


def normalise_number(number: Union[list, tuple, str]) -> List[int]:
    """Normalises the input number to a list of integers.

    Parameters
    ----------
    number : Union[list, tuple, str]
        The number to normalise.

    Returns
    -------
    list
        The normalised list of integers.
    """
    number = to_list(number)

    if len(number) > 4 and (number[1] == "to" and number[3] == "by"):
        return list(range(int(number[0]), int(number[2]) + 1, int(number[4])))

    if len(number) > 2 and number[1] == "to":
        return list(range(int(number[0]), int(number[2]) + 1))

    return number


def normalise_request(request: Dict) -> Dict:
    """Normalises the request dictionary by converting certain fields to lists.

    Parameters
    ----------
    request : dict
        The request dictionary to normalise.

    Returns
    -------
    dict
        The normalised request dictionary.
    """
    request = deepcopy(request)
    if "number" in request:
        request["number"] = normalise_number(request["number"])
    if "time" in request:
        request["time"] = to_list(request["time"])
    request["param"] = to_list(request["param"])
    return request


def load_if_needed(context: Any, dates: Any, dict_or_dataset: Union[Dict, Any]) -> Any:
    """Loads the dataset if the input is a dictionary, otherwise returns the input.

    Parameters
    ----------
    context : Any
        The context for loading the dataset.
    dates : Any
        The dates for loading the dataset.
    dict_or_dataset : Union[dict, Any]
        The input dictionary or dataset.

    Returns
    -------
    Any
        The loaded dataset or the original input.
    """
    if isinstance(dict_or_dataset, dict):
        dict_or_dataset = normalise_request(dict_or_dataset)
        dict_or_dataset = mars(context, dates, dict_or_dataset)
    return dict_or_dataset


@legacy_source(__file__)
def recentre(
    context: Any,
    dates: Any,
    members: Union[Dict, Any],
    centre: Union[Dict, Any],
    alpha: float = 1.0,
    remapping: Dict = {},
    patches: Dict = {},
) -> Any:
    """Recentres the members dataset using the centre dataset.

    Parameters
    ----------
    context : Any
        The context for recentering.
    dates : Any
        The dates for recentering.
    members : Union[dict, Any]
        The members dataset or request dictionary.
    centre : Union[dict, Any]
        The centre dataset or request dictionary.
    alpha : float, optional
        The alpha value for recentering. Defaults to 1.0.
    remapping : dict, optional
        The remapping dictionary. Defaults to {}.
    patches : dict, optional
        The patches dictionary. Defaults to {}.

    Returns
    -------
    Any
        The recentred dataset.
    """
    members = load_if_needed(context, dates, members)
    centre = load_if_needed(context, dates, centre)
    return _recentre(members=members, centre=centre, alpha=alpha)


execute = recentre
